# Natural Language Toolkit: Sourced Strings
#
# Copyright (C) 2001-2009 NLTK Project
# Author: Edward Loper <edloper@gmail.com>
# URL: <http://www.nltk.org/>
# For license information, see LICENSE.TXT

"""
"Sourced strings" are strings that are annotated with information
about the location in a document where they were originally found.
Sourced strings are subclassed from Python strings.  As a result, they
can usually be used anywhere a normal Python string can be used.

    >>> from nltk.sourcedstring import SourcedString
    >>> newt_contents = '''
    ... She turned me into a newt!
    ... I got better.'''.strip()
    >>> newt_doc = SourcedString(newt_contents, 'newt.txt')
    >>> newt_doc
    'She turned me into a newt!\\nI got better.'@[0:40]
    >>> newt = newt_doc.split()[5] # Find the sixth word.
    >>> newt
    'newt!'@[21:26]

"""

import re, sys
from nltk.internals import slice_bounds, abstract

__all__ = [
    'StringSource',
    'ConsecutiveCharStringSource', 'ContiguousCharStringSource',
    'SourcedString', 'SourcedStringStream', 'SourcedStringRegexp',
    'SimpleSourcedString', 'CompoundSourcedString',
    'SimpleSourcedByteString', 'SimpleSourcedUnicodeString',
    'CompoundSourcedByteString', 'CompoundSourcedUnicodeString',
    ]

#//////////////////////////////////////////////////////////////////////
# String Sources
#//////////////////////////////////////////////////////////////////////

class StringSource(object):
    """
    A description of the location of a string in a document.  Each
    ``StringSource`` consists of a document identifier, along with
    information about the begin and end offsets of each character in
    the string.  These offsets are typically either byte offsets or
    character offsets.  (Note that for unicode strings, byte offsets
    and character offsets are not the same thing.)

    ``StringSource`` is an abstract base class.  Two concrete
    subclasses are used depending on the properties of the string
    whose source is being described:

      - ``ConsecutiveCharStringSource`` describes the source of strings
        whose characters have consecutive offsets (in particular, byte
        strings w/ byte offsets; and unicode strings with character
        offsets).

      - ``ContiguousCharStringSource`` describes the source of strings
        whose characters are contiguous, but do not necessarily have
        consecutive offsets (in particular, unicode strings with byte
        offsets).

    :ivar docid: An identifier (such as a filename) that specifies
        which document contains the string.

    :ivar offsets: A list of offsets specifying the location of each
        character in the document.  The *i* th character of the string
        begins at offset ``offsets[i]`` and ends at offset
        ``offsets[i+1]``.  The length of the ``offsets`` list is one
        greater than the list of the string described by this
        ``StringSource``.

    :ivar begin: The document offset where the string begins.  (I.e.,
        the offset of the first character in the string.)
        ``source.begin`` is always equal to ``source.offsets[0]``.

    :ivar end: The document offset where the string ends.  (For
        character offsets, one plus the offset of the last character;
        for byte offsets, one plus the offset of the last byte that
        encodes the last character).  ``source.end`` is always equal
        to ``source.offsets[-1]``.
    """
    def __new__(cls, docid, *args, **kwargs):
        # If the StringSource constructor is called directly, then
        # choose one of its subclasses to delegate to.
        if cls is StringSource:
            if args:
                raise TypeError("Specifcy either begin and end, or "
                                 "offsets, using keyword arguments")
            if 'begin' in kwargs and 'end' in kwargs and 'offsets' not in kwargs:
                cls = ConsecutiveCharStringSource
            elif ('begin' not in kwargs and 'end' not in kwargs and
                  'offsets' in kwargs):
                cls = ContiguousCharStringSource
            else:
                raise TypeError("Specify either begin and end, or offsets "
                                 "(but not both)")
        # Construct the object.
        return object.__new__(cls)

    def __init__(self, docid, **kwargs):
        """
        Create a new ``StringSource``.  When the ``StringSource``
        constructor is called directly, it automatically delegates to
        one of its two subclasses:

            - If ``begin`` and ``end`` are specified, then a
              ``ConsecutiveCharStringSource`` is returned.
            - If ``offsets`` is specified, then a
              ``ContiguousCharStringSource`` is returned.

        In both cases, the arguments must be specified as keyword
        arguments (not positional arguments).
        """

    def __getitem__(self, index):
        """
        Return a ``StringSource`` describing the location where the
        specified character was found.  In particular, if ``s`` is the
        string that this source describes, then return a
        ``StringSource`` describing the location of ``s[index]``.

        :raise IndexError: If index is out of range.
        """
        if isinstance(index, slice):
            start, stop = slice_bounds(self, index)
            return self.__getslice__(start, stop)
        else:
            if index < 0: index += len(self)
            if index < 0 or index >= len(self):
                raise IndexError('StringSource index out of range')
            return self.__getslice__(index, index+1)

    @abstract
    def __getslice__(self, start, stop):
        """
        Return a ``StringSource`` describing the location where the
        specified substring was found.  In particular, if ``s`` is the
        string that this source describes, then return a
        ``StringSource`` describing the location of ``s[start:stop]``.
        """

    @abstract
    def __len__(self):
        """
        Return the length of the string described by this
        ``StringSource``.  Note that this may not be equal to
        ``self.end-self.begin`` for unicode strings described using
        byte offsets.
        """

    def __str__(self):
        if self.end == self.begin+1:
            return '@%s[%s]' % (self.docid, self.begin,)
        else:
            return '@%s[%s:%s]' % (self.docid, self.begin, self.end)

    def __cmp__(self, other):
        return (cmp(self.docid, self.docid) or
                cmp([(charloc.begin, charloc.end) for charloc in self],
                    [(charloc.begin, charloc.end) for charloc in other]))

    def __hash__(self):
        # Cache hash values.
        if not hasattr(self, '_hash'):
            self._hash = hash( (self.docid,
                                tuple((charloc.begin, charloc.end)
                                      for charloc in self)) )
        return self._hash

class ConsecutiveCharStringSource(StringSource):
    """
    A ``StringSource`` that specifies the source of strings whose
    characters have consecutive offsets.  In particular, the following
    two properties must hold for all valid indices:

      - source[i].end == source[i].begin + 1
      - source[i].end == source[i+1].begin

    These properties allow the source to be stored using just a start
    offset and an end offset (along with a docid).

    This ``StringSource`` can be used to describe byte strings that are
    indexed using byte offsets or character offsets; or unicode
    strings that are indexed using character offsets.
    """
    def __init__(self, docid, begin, end):
        if not isinstance(begin, (int, long)):
            raise TypeError("begin attribute expected an integer")
        if not isinstance(end, (int, long)):
            raise TypeError("end attribute expected an integer")
        if not  end >= begin:
            raise ValueError("begin must be less than or equal to end")
        self.docid = docid
        self.begin = begin
        self.end = end

    @property
    def offsets(self):
        return tuple(range(self.begin, self.end+1))

    def __len__(self):
        return self.end-self.begin

    def __getslice__(self, start, stop):
        start = max(0, min(len(self), start))
        stop = max(start, min(len(self), stop))
        return ConsecutiveCharStringSource(
            self.docid, self.begin+start, self.begin+stop)

    def __cmp__(self, other):
        if isinstance(other, ConsecutiveCharStringSource):
            return (cmp(self.docid, other.docid) or
                    cmp(self.begin, other.begin) or
                    cmp(self.end, other.end))
        else:
            return StringSource.__cmp__(self, other)

    def __repr__(self):
        return 'StringSource(%r, begin=%r, end=%r)' % (
            self.docid, self.begin, self.end)

class ContiguousCharStringSource(StringSource):
    """
    A ``StringSource`` that specifies the source of strings whose
    character are contiguous, but do not necessarily have consecutive
    offsets.  In particular, each character's end offset must be equal
    to the next character's start offset:

      - source[i].end == source[i+1].begin

    This property allow the source to be stored using a list of
    ``len(source)+1`` offsets (along with a docid).

    This ``StringSource`` can be used to describe unicode strings that
    are indexed using byte offsets.
    """
    CONSTRUCTOR_CHECKS_OFFSETS = False
    def __init__(self, docid, offsets):
        offsets = tuple(offsets)
        if len(offsets) == 0:
            raise ValueError("at least one offset must be specified")
        if self.CONSTRUCTOR_CHECKS_OFFSETS:
            for i in range(len(offsets)):
                if not isinstance(offsets[i], (int,long)):
                    raise TypeError("offsets must be integers")
                if i>0 and offsets[i-1]>offsets[i]:
                    raise TypeError("offsets must be monotonic increasing")
        self.docid = docid
        self.offsets = offsets

    @property
    def begin(self): return self.offsets[0]

    @property
    def end(self): return self.offsets[-1]

    def __len__(self):
        return len(self.offsets)-1

    def __getslice__(self, start, stop):
        start = max(0, min(len(self), start))
        stop = max(start, min(len(self), stop))
        return ContiguousCharStringSource(
            self.docid, self.offsets[start:stop+1])

    def __cmp__(self, other):
        if isinstance(other, ConsecutiveCharStringSource):
            return (cmp(self.docid, other.docid) or
                    cmp(self.offsets, other._offsets))
        else:
            return StringSource.__cmp__(self, other)

    def __repr__(self):
        return 'StringSource(%r, offsets=%r)' % (self.docid, self.offsets)

#//////////////////////////////////////////////////////////////////////
# Base Class for Sourced Strings.
#//////////////////////////////////////////////////////////////////////

class SourcedString(object):
    """
    A string that is annotated with information about the location in
    a document where it was originally found.  Sourced strings are
    subclassed from Python strings.  As a result, they can usually be
    used anywhere a normal Python string can be used.

    There are two types of sourced strings: ``SimpleSourcedString``s,
    which correspond to a single substring of a document; and
    ``CompoundSourcedString``s, which are constructed by concatenating
    strings from multiple sources.  Each of these types has two
    concrete subclasses: one for unicode strings (subclassed from
    ``unicode``), and one for byte strings (subclassed from ``str``).

    Two sourced strings are considered equal if their contents are
    equal, even if their sources differ.  This fact is important in
    ensuring that sourced strings act like normal strings.  In
    particular, it allows sourced strings to be used with code that
    was originally intended to process plain Python strings.

    If you wish to determine whether two sourced strings came from the
    same location in the same document, simply compare their
    ``sources`` attributes.  If you know that both sourced strings are
    ``SimpleSourcedStrings``, then you can compare their ``source``
    attribute instead.

    String operations that act on sourced strings will preserve
    location information whenever possible.  However, there are a few
    types of string manipulation that can cause source information to
    be discarded.  The most common examples of operations that will
    lose source information are:

      - ``str.join()``, where the joining string is not sourced.
      - ``str.replace()``, where the original string is not sourced.
      - String formatting (the ``%`` operator).
      - Regular expression substitution.

    :ivar sources: A sorted tuple of ``(index, source)`` pairs.  Each
        such pair specifies that the source of
        ``self[index:index+len(source)]`` is ``source``.  Any characters
        for which no source is specified are sourceless (e.g., plain
        Python characters that were concatenated to a sourced string).

        When working with simple sourced strings, it's usually easier
        to use the ``source`` attribute instead; however, the
        ``sources`` attribute is defined for both simple and compound
        sourced strings.
    """
    def __new__(cls, contents, source):
        # If the SourcedString constructor is called directly, then
        # choose one of its subclasses to delegate to.
        if cls is SourcedString:
            if isinstance(contents, str):
                cls = SimpleSourcedByteString
            elif isinstance(contents, unicode):
                cls = SimpleSourcedUnicodeString
            else:
                raise TypeError("Expected 'contents' to be a unicode "
                                "string or a byte string")

        # Create the new object using the appropriate string class's
        # __new__, which takes just the contents argument.
        return cls._stringtype.__new__(cls, contents)

    _stringtype = None
    """A class variable, defined by subclasses of ``SourcedString``,
       determining what type of string this class contains.  Its
       value must be either str or ``unicode``."""

    #//////////////////////////////////////////////////////////////////////
    #{ Splitting & Stripping Methods
    #//////////////////////////////////////////////////////////////////////

    def lstrip(self, chars=None):
        s = self._stringtype.lstrip(self, chars)
        return self[len(self)-len(s):]

    def rstrip(self, chars=None):
        s = self._stringtype.rstrip(self, chars)
        return self[:len(s)]

    def strip(self, chars=None):
        return self.lstrip(chars).rstrip(chars)

    _WHITESPACE_RE = re.compile(r'\s+')
    def split(self, sep=None, maxsplit=None):
        # Check for unicode/bytestring mismatches:
        if self._mixed_string_types(sep, maxsplit):
            return self._decode_and_call('split', sep, maxsplit)
        # Use a regexp to split self.
        if sep is None: sep_re = self._WHITESPACE_RE
        else: sep_re = re.compile(re.escape(sep))
        if maxsplit is None: return sep_re.split(self)
        else: return sep_re.split(self, maxsplit)

    def rsplit(self, sep=None, maxsplit=None):
        # Check for unicode/bytestring mismatches:
        if self._mixed_string_types(sep, maxsplit):
            return self._decode_and_call('rsplit', sep, maxsplit)
        # Split on whitespace use a regexp.
        if sep is None:
            seps = list(self._WHITESPACE_RE.finditer(self))
            if maxsplit: seps = seps[-maxsplit:]
            if not seps: return [self]
            result = [self[:seps[0].start()]]
            for i in range(1, len(seps)):
                result.append(self[seps[i-1].end():seps[i].start()])
            result.append(self[seps[-1].end():])
            return result
        # Split on a given string: use rfind.
        else:
            result = []
            piece_end = len(self)
            while maxsplit != 0:
                sep_pos = self.rfind(sep, 0, piece_end)
                if sep_pos < 0: break
                result.append(self[sep_pos+len(sep):piece_end])
                piece_end = sep_pos
                if maxsplit is not None: maxsplit -= 1
            if piece_end > 0:
                result.append(self[:piece_end])
            return result[::-1]

    def partition(self, sep):
        head, sep, tail = self._stringtype.partition(self, sep)
        i, j = len(head), len(head)+len(sep)
        return (self[:i], self[i:j], self[j:])

    def rpartition(self, sep):
        head, sep, tail = self._stringtype.rpartition(self, sep)
        i, j = len(head), len(head)+len(sep)
        return (self[:i], self[i:j], self[j:])

    _NEWLINE_RE = re.compile(r'\n')
    _LINE_RE = re.compile(r'.*\n?')
    def splitlines(self, keepends=False):
        if keepends:
            return self._LINE_RE.findall(self)
        else:
            return self._NEWLINE_RE.split(self)

    #//////////////////////////////////////////////////////////////////////
    #{ String Concatenation Methods
    #//////////////////////////////////////////////////////////////////////

    @staticmethod
    def concat(substrings):
        """
        Return a sourced string formed by concatenating the given list
        of substrings.  Adjacent substrings will be merged when
        possible.

        Depending on the types and values of the supplied substrings,
        the concatenated string's value may be a Python string (str
        or ``unicode``), a ``SimpleSourcedString``, or a
        ``CompoundSourcedString``.
        """
        # Flatten nested compound sourced strings, and merge adjacent
        # strings where possible:
        merged = []
        for substring in substrings:
            SourcedString.__add_substring_to_list(substring, merged)

        # Return the concatenated string.
        if len(merged) == 0:
            return ''
        elif len(merged) == 1:
            return merged[0]
        else:
            return CompoundSourcedString(merged)

    def __add__(self, other):
        return SourcedString.concat([self, other])

    def __radd__(self, other):
        return SourcedString.concat([other, self])

    def __mul__(self, other):
        if other <= 0:
            return self._stringtype('')
        else:
            result = self
            for i in range(1, other):
                result += self
            return result

    def __rmul__(self, other):
        return self.__mul__(other)

    def join(self, sequence):
        seq_iter = iter(sequence)
        # Add the first element; but if sequence is empty, return an
        # empty string.
        try:
            s = seq_iter.next()
        except StopIteration:
            return self._stringtype('')

        # Add the remaining elements, separated by self.
        for elt in seq_iter:
            s += self
            s += elt
        return s

    @staticmethod
    def __add_substring_to_list(substring, result):
        """
        Helper for ``concat()``: add ``substring`` to the end of the
        list of substrings in ``result``.  If ``substring`` is compound,
        then add its own substrings instead.  Merge adjacent
        substrings whenever possible.  Discard empty un-sourced
        substrings.
        """
        # Flatten nested compound sourced strings.
        if isinstance(substring, CompoundSourcedString):
            for s in substring.substrings:
                SourcedString.__add_substring_to_list(s, result)

        # Discard empty Python substrings.
        elif len(substring) == 0 and not isinstance(substring, SourcedString):
            pass # discard.

        # Merge adjacent simple sourced strings (when possible).
        elif (result and isinstance(result[-1], SimpleSourcedString) and
              isinstance(substring, SimpleSourcedString) and
              result[-1].end == substring.begin and
              result[-1].docid == substring.docid):
            result[-1] = SourcedString.__merge_simple_substrings(
                result[-1], substring)

        # Merge adjacent Python strings.
        elif (result and not isinstance(result[-1], SourcedString) and
              not isinstance(substring, SourcedString)):
            result[-1] += substring

        # All other strings just get appended to the result list.
        else:
            result.append(substring)

    @staticmethod
    def __merge_simple_substrings(lhs, rhs):
        """
        Helper for ``__add_substring_to_list()``: Merge ``lhs`` and
        ``rhs`` into a single simple sourced string, and return it.
        """
        contents = lhs._stringtype.__add__(lhs, rhs)
        if (isinstance(lhs.source, ConsecutiveCharStringSource) and
            isinstance(rhs.source, ConsecutiveCharStringSource)):
            source = ConsecutiveCharStringSource(
                lhs.source.docid, lhs.source.begin, rhs.source.end)
        else:
            source = ContiguousCharStringSource(
                lhs.source.docid, lhs.source.offsets+rhs.source.offsets[1:])
        return SourcedString(contents, source)

    #//////////////////////////////////////////////////////////////////////
    #{ Justification Methods
    #//////////////////////////////////////////////////////////////////////

    def center(self, width, fillchar=' '):
        return (fillchar * ((width-len(self))/2)  + self +
                fillchar * ((width-len(self)+1)/2))

    def ljust(self, width, fillchar=' '):
        return self + fillchar * (width-len(self))

    def rjust(self, width, fillchar=' '):
        return fillchar * (width-len(self)) + self

    def zfill(self, width):
        return self.rjust(width, '0')

    #//////////////////////////////////////////////////////////////////////
    #{ Replacement Methods
    #//////////////////////////////////////////////////////////////////////

    # [xx] There's no reason in principle why this can't preserve
    # location information.  But for now, it doesn't.
    def __mod__(self, other):
        return self._stringtype.__mod__(self, other)

    def replace(self, old, new, count=0):
        # Check for unicode/bytestring mismatches:
        if self._mixed_string_types(old, new, count):
            return self._decode_and_call('replace', old, new, count)
        # Use a regexp to find all occurrences of old, and replace them w/ new.
        result = ''
        pos = 0
        for match in re.finditer(re.escape(old), self):
            result += self[pos:match.start()]
            result += new
            pos = match.end()
        result += self[pos:]
        return result

    def expandtabs(self, tabsize=8):
        if len(self) == 0: return self
        pieces = re.split(r'([\t\n])', self)
        result = ''
        offset = 0
        for piece in pieces:
            if piece == '\t':
                spaces = 8 - (offset % tabsize)
                # Each inserted space's source is the same as the
                # source of the tab character that generated it.
                result += spaces * SourcedString(' ', piece.source)
                offset = 0
            else:
                result += piece
                if piece == '\n': offset = 0
                else: offset += len(piece)
        return result

    def translate(self, table, deletechars=''):
        # Note: str.translate() and unicode.translate() have
        # different interfaces.
        if isinstance(self, unicode):
            if deletechars:
                raise TypeError('The unicode version of translate() does not '
                                'accept the deletechars parameter')
            return SourcedString.concat(
                [SourcedString(table.get(c,c), c.source)
                 for c in self if table.get(c,c) is not None])
        else:
            if len(table) != 256:
                raise ValueError('translation table must be 256 characters long')
            return SourcedString.concat(
                [SourcedString(table[ord(c)], c.source)
                 for c in self if c not in deletechars])

    #//////////////////////////////////////////////////////////////////////
    #{ Unicode
    #//////////////////////////////////////////////////////////////////////

    # Unicode string -> byte string
    def encode(self, encoding=None, errors='strict'):
        if encoding is None: encoding = sys.getdefaultencoding()
        if isinstance(self, str):
            return self.decode().encode(encoding, errors)

        # Encode characters one at a time.
        result = []
        for i, char in enumerate(self):
            char_bytes = self._stringtype.encode(char, encoding, errors)
            for char_byte in char_bytes:
                if isinstance(char, SimpleSourcedString):
                    result.append(SourcedString(char_byte, char.source))
                else:
                    assert not isinstance(char, CompoundSourcedString)
                    result.append(char_byte)
        return SourcedString.concat(result)

    # Byte string -> unicode string.
    def decode(self, encoding=None, errors='strict'):
        if encoding is None: encoding = sys.getdefaultencoding()
        if isinstance(self, unicode):
            return self.encode().decode(encoding, errors)

        # Decode self into a plain unicode string.
        unicode_chars = self._stringtype.decode(self, encoding, errors)

        # Special case: if the resulting string has the same length
        # that the source string does, then we can safely assume that
        # each character is encoded with one byte; so we can just
        # reuse our source.
        if len(unicode_chars) == len(self):
            return self._decode_one_to_one(unicode_chars)

        # Otherwise: re-encode the characters, one at a time, to
        # determine how long their encodings are.
        result = []
        first_byte = 0
        for unicode_char in unicode_chars:
            char_width = len(unicode_char.encode(encoding, errors))
            last_byte = first_byte + char_width - 1
            if (isinstance(self[first_byte], SourcedString) and
                isinstance(self[last_byte], SourcedString)):
                begin = self[first_byte].begin
                end = self[last_byte].end
                if end-begin == 1:
                    source = StringSource(docid=self[first_byte].docid,
                                          begin=begin, end=end)
                else:
                    source = StringSource(docid=self[first_byte].docid,
                                          offsets=[begin, end])
                result.append(SourcedString(unicode_char, source))
            else:
                result.append(unicode_char)

            # First byte of the next char is 1+last byte of this char.
            first_byte = last_byte+1
        if last_byte+1 != len(self):
            raise AssertionError("SourcedString.decode() does not support "
                                 "encodings that are not symmetric.")

        return SourcedString.concat(result)

    @abstract
    def _decode_one_to_one(unicode_chars):
        """
        Helper for ``self.decode()``.  Returns a unicode-decoded
        version of this ``SourcedString``.  ``unicode_chars`` is the
        unicode-decoded contents of this ``SourcedString``.

        This is used in the special case where the decoded string has
        the same length that the source string does.  As a result, we
        can safely assume that each character is encoded with one
        byte; so we can just reuse our source.  E.g., this will happen
        when decoding an ASCII string with utf-8.
        """

    def _mixed_string_types(self, *args):
        """
        Return true if the list (self,)+args contains at least one
        unicode string and at least one byte string.  (If this is the
        case, then all byte strings should be converted to unicode by
        calling decode() before the operation is performed.  You can
        do this automatically using ``_decode_and_call()``.
        """
        any_unicode = isinstance(self, unicode)
        any_bytestring = isinstance(self, str)
        for arg in args:
            any_unicode = any_unicode or isinstance(arg, unicode)
            any_bytestring = any_bytestring or isinstance(arg, str)
        return any_unicode and any_bytestring

    def _decode_and_call(self, op, *args):
        """
        If self or any of the values in args is a byte string, then
        convert it to unicode by calling its decode() method.  Then
        return the result of calling self.op(*args).  ``op`` is
        specified using a string, because if ``self`` is a byte string,
        then it will change type when it is decoded.
        """
        # Make sure all args are decoded to unicode.
        args = list(args)
        for i in range(len(args)):
            if isinstance(args[i], str):
                args[i] = args[i].decode()
        # Make sure self is decoded to unicode.
        if isinstance(self, str):
            self = self.decode()
        # Retry the operation.
        method = getattr(self, op)
        return method(*args)

    #//////////////////////////////////////////////////////////////////////
    #{ Display
    #//////////////////////////////////////////////////////////////////////

    def pprint(self, vertical=False, wrap=70):
        """
        Return a string containing a pretty-printed display of this
        sourced string.

        :param vertical: If true, then the returned display string will
            have vertical orientation, rather than the default horizontal
            orientation.

        :param wrap: Controls when the pretty-printed output is wrapped
            to the next line.  If ``wrap`` is an integer, then lines are
            wrapped when they become longer than ``wrap``.  If ``wrap`` is
            a string, then lines are wrapped immediately following that
            string.  If ``wrap`` is None, then lines are never wrapped.
        """
        if len(self) == 0: return '[Empty String]'
        if vertical == 1: return self._pprint_vertical() # special-cased

        max_digits = len(str(max(max(getattr(c, 'begin', 0),
                                     getattr(c, 'end', 0)) for c in self)))
        if not isinstance(wrap, (basestring, int, long, type(None))):
            raise TypeError("Expected wrap to be a sring, int, or None.")

        result = []
        prev_offset = None # most recently displayed offset.
        prev_docid = None
        docid_line = ''
        output_lines = [''] * (max_digits+2)

        for pos, char in enumerate(self):
            char_begin = getattr(char, 'begin', None)
            char_end = getattr(char, 'end', None)
            char_docid = getattr(char, 'docid', None)

            # If the docid changed, then display the docid for the
            # previous segment.
            if char_docid != prev_docid:
                width = len(output_lines[0]) - len(docid_line)
                docid_line += self._pprint_docid(width, prev_docid)
                prev_docid = char_docid

            # Put a cap on the beginning of sourceless strings
            elif not output_lines[0] and char_begin is None:
                self._pprint_offset(' ', output_lines)

            # Display the character.
            if char_begin != prev_offset:
                self._pprint_offset(char_begin, output_lines)
            self._pprint_char(char, output_lines)
            self._pprint_offset(char_end, output_lines)
            prev_offset = char_end

            # Decide whether we're at the end of the line or not.
            line_len = len(output_lines[0])
            if ( (isinstance(wrap, basestring) and
                  self[max(0,pos-len(wrap)+1):pos+1] == wrap) or
                 (isinstance(wrap, (int,long)) and line_len>=wrap) or
                 pos == len(self)-1):

                # Put a cap on the end of sourceless strings
                if char_end is None:
                    self._pprint_offset(' ', output_lines)

                # Filter out any empty output lines.
                output_lines = [l for l in output_lines if l.strip()]

                # Draw the docid line
                width = len(output_lines[0]) - len(docid_line)
                docid_line += self._pprint_docid(width, prev_docid)
                result.append(docid_line)

                # Draw the output lines
                for output_line in reversed(output_lines):
                    result.append(output_line)
                result.append(output_lines[1])

                # Reset variables for the next line.
                prev_offset = None
                prev_docid = None
                docid_line = ''
                output_lines = [''] * (max_digits+2)

        return '\n'.join(result)

    def _pprint_vertical(self):
        result = []
        prev_offset = None
        max_digits = len(str(max(max(getattr(c, 'begin', 0),
                                     getattr(c, 'end', 0)) for c in self)))
        for pos, char in enumerate(self):
            char_begin = getattr(char, 'begin', None)
            char_end = getattr(char, 'end', None)
            char_docid = getattr(char, 'docid', None)

            if char_begin is None:
                assert char_end is None
                if pos == 0: result.append('+-----+')
                result.append(':%s:' %
                              self._pprint_char_repr(char).center(5))
                if pos == len(self)-1: result.append('+-----+')
                prev_offset = None
            else:
                if char_begin != prev_offset:
                    result.append('+-----+ %s [%s]' % (
                        str(char_begin).rjust(max_digits), char_docid))
                result.append('|%s| %s [%s]' % (
                    self._pprint_char_repr(char).center(5),
                    ' '*max_digits, char_docid))
                result.append('+-----+ %s [%s]' % (
                    str(char_end).rjust(max_digits), char_docid))
            prev_offset = char_end
        return '\n'.join(result)

    _PPRINT_CHAR_REPRS = {'\n': r'\n', '\r': r'\r',
                          '\a': r'\a', '\t': r'\t'}

    def _pprint_docid(self, width, docid):
        if docid is None: return ' '*width
        else: return '[%s]' % (docid[:width-2].center(width-2, '='))

    def _pprint_char_repr(self, char):
        # Decide how to represent this character.
        if 32 <= ord(char) <= 127:
            return str(char)
        elif char in self._PPRINT_CHAR_REPRS:
            return self._PPRINT_CHAR_REPRS[char]
        elif isinstance(char, str):
            return r'\x%02x' % ord(char)
        else:
            return r'\u%04x' % ord(char)

    def _pprint_char(self, char, output_lines):
        """Helper for ``pprint()``: add a character to the
        pretty-printed output."""
        char_repr = self._pprint_char_repr(char)
        output_lines[0] += char_repr
        # Add fillers to the offset lines.
        output_lines[1] += '-'*len(char_repr)
        for i in range(2, len(output_lines)):
            output_lines[i] += ' '*len(char_repr)

    def _pprint_offset(self, offset, output_lines):
        """Helper for ``pprint()``: add an offset marker to the
        pretty-printed output."""
        if offset is None: return
        output_lines[0] += '|'
        output_lines[1] += '+'
        offset_rep = str(offset).rjust(len(output_lines)-2)
        for digit in range(len(offset_rep)):
            output_lines[-digit-1] += offset_rep[digit]

#//////////////////////////////////////////////////////////////////////
# Simple Sourced String
#//////////////////////////////////////////////////////////////////////

class SimpleSourcedString(SourcedString):
    """
    A single substring of a document, annotated with information about
    the location in the document where it was originally found.  See
    ``SourcedString`` for more information.
    """
    def __new__(cls, contents, source):
        # If the SimpleSourcedString constructor is called directly,
        # then choose one of its subclasses to delegate to.
        if cls is SimpleSourcedString:
            if isinstance(contents, str):
                cls = SimpleSourcedByteString
            elif isinstance(contents, unicode):
                cls = SimpleSourcedUnicodeString
            else:
                raise TypeError("Expected 'contents' to be a unicode "
                                "string or a byte string")

        # Create the new object using the appropriate string class's
        # __new__, which takes just the contents argument.
        return cls._stringtype.__new__(cls, contents)

    def __init__(self, contents, source):
        """
        Construct a new sourced string.

        :param contents: The string contents of the new sourced string.
        :type contents: str or unicode
        :param source: The source for the new string.  If ``source`` is
            a string, then it is used to automatically construct a new
            ``ConsecutiveCharStringSource`` with a begin offset of
            ``0`` and an end offset of ``len(contents)``.  Otherwise,
            ``source`` shoulde be a ``StringSource`` whose length matches
            the length of ``contents``.
        """
        if not isinstance(source, StringSource):
            source = ConsecutiveCharStringSource(source, 0, len(contents))
        elif len(source) != len(contents):
            raise ValueError("Length of source (%d) must match length of "
                             "contents (%d)" % (len(source), len(contents)))

        self.source = source
        """A ``StringLocation`` specifying the location where this string
           occurred in the source document."""

    @property
    def begin(self):
        """
        The document offset where the string begins.  (I.e.,
        the offset of the first character in the string.)"""
        return self.source.begin

    @property
    def end(self):
        """The document offset where the string ends.  (For character
        offsets, one plus the offset of the last character; for byte
        offsets, one plus the offset of the last byte that encodes the
        last character)."""
        return self.source.end

    @property
    def docid(self):
        """
        An identifier (such as a filename) that specifies the document
        where the string was found.
        """
        return self.source.docid

    @property
    def sources(self):
        return ((0, self.source),)

    def __repr__(self):
        if self.end == self.begin+1:
            source_repr =  '@[%s]' % (self.begin,)
        else:
            source_repr = '@[%s:%s]' % (self.begin, self.end)
        return self._stringtype.__repr__(self) + source_repr

    def __getitem__(self, index):
        result = self._stringtype.__getitem__(self, index)
        if isinstance(index, slice):
            if index.step not in (None, 1):
                return result
            else:
                start, stop = slice_bounds(self, index)
                return self.__getslice__(start, stop)
        else:
            return SourcedString(result, self.source[index])

    def __getslice__(self, start, stop):
        # Negative indices get handled *before* __getslice__ is
        # called.  Restrict start/stop to be within the range of the
        # string, to prevent negative indices from being adjusted
        # twice.
        start = max(0, min(len(self), start))
        stop = max(start, min(len(self), stop))

        return SourcedString(
            self._stringtype.__getslice__(self, start, stop),
            self.source[start:stop])

    def capitalize(self):
        result = self._stringtype.capitalize(self)
        return SourcedString(result, self.source)

    def lower(self):
        result = self._stringtype.lower(self)
        return SourcedString(result, self.source)

    def upper(self):
        result = self._stringtype.upper(self)
        return SourcedString(result, self.source)

    def swapcase(self):
        result = self._stringtype.swapcase(self)
        return SourcedString(result, self.source)

    def title(self):
        result = self._stringtype.title(self)
        return SourcedString(result, self.source)

    def _decode_one_to_one(self, unicode_chars):
        return SourcedString(unicode_chars, self.source)

#//////////////////////////////////////////////////////////////////////
# Compound Sourced String
#//////////////////////////////////////////////////////////////////////

class CompoundSourcedString(SourcedString):
    """
    A string constructed by concatenating substrings from multiple
    sources, and annotated with information about the locations where
    those substrings were originally found.  See ``SourcedString`` for
    more information.

    :ivar substrings: The tuple of substrings that compose this
        compound sourced string.  Every compound sourced string is
        required to have at least two substrings; and the substrings
        themselves may never be CompoundSourcedStrings.
    """
    def __new__(cls, substrings):
        # If the CompoundSourcedString constructor is called directly,
        # then choose one of its subclasses to delegate to.
        if cls is CompoundSourcedString:
            # Decide whether to use a unicode string or a byte string.
            use_unicode = sum(1 for substring in substrings
                              if isinstance(substring, unicode))
            if use_unicode:
                cls = CompoundSourcedUnicodeString
            else:
                cls = CompoundSourcedByteString

        # Build the concatenated string using str.join(), which will
        # return a str or unicode object; never a sourced string.
        contents = ''.join(substrings)

        # Create the new object using the appropriate string class's
        # __new__, which takes just the contents argument.
        return cls._stringtype.__new__(cls, contents)

    def __init__(self, substrings):
        """
        Construct a new compound sourced string that combines the
        given list of substrings.

        Typically, compound sourced strings should not be constructed
        directly; instead, use ``SourcedString.concat()``, which
        flattens nested compound sourced strings, and merges adjacent
        substrings when possible.

        :raise ValueError: If ``len(substrings)  < 2``
        :raise ValueError: If ``substrings`` contains any
            ``CompoundSourcedString``s.
        """
        if len(substrings) < 2:
            raise ValueError("CompoundSourcedString requires at least "
                             "two substrings")

        # Don't nest compound sourced strings.
        for substring in substrings:
            if isinstance(substring, CompoundSourcedString):
                raise ValueError("substrings may not contain "
                                 "CompoundSourcedStrings.")

        self.substrings = tuple(substrings)

    @property
    def sources(self):
        index = 0
        source_list = []
        for substring in self.substrings:
            if isinstance(substring, SourcedString):
                source_list.append( (index, substring.source) )
            index += len(substring)
        return tuple(source_list)

    def __repr__(self):
        sources = [self._source_repr(s) for s in self.substrings]
        source_str = '@[%s]' % ','.join(sources)
        return self._stringtype.__repr__(self) + source_str

    def _source_repr(self, substring):
        if isinstance(substring, SimpleSourcedString):
            return '%s:%s' % (substring.begin, substring.end)
        else:
            return '...'

    def __getitem__(self, index):
        if isinstance(index, slice):
            if index.step not in (None, 1):
                return self._stringtype.__getitem__(self, index)
            else:
                start, stop = slice_bounds(self, index)
                return self.__getslice__(start, stop)
        else:
            if index < 0: index += len(self)
            if index < 0 or index >= len(self):
                raise IndexError('StringSource index out of range')
            return self.__getslice__(index, index+1)

    def __getslice__(self, start, stop):
        # Bounds checking.
        start = max(0, min(len(self), start))
        stop = max(start, min(len(self), stop))

        # Construct a source list for the resulting string.
        result_substrings = []
        offset = 0
        for substring in self.substrings:
            if offset+len(substring) > start:
                s, e = max(0, start-offset), stop-offset
                result_substrings.append(substring[s:e])
            offset += len(substring)
            if offset >= stop: break

        # Concatentate the resulting substrings.
        if len(result_substrings) == 0:
            return ''
        elif len(result_substrings) == 1:
            return result_substrings[0]
        else:
            return SourcedString.concat(result_substrings)

    def capitalize(self):
        return SourcedString.concat([s.capitalize() for s in self.substrings])

    def lower(self):
        return SourcedString.concat([s.lower() for s in self.substrings])

    def upper(self):
        return SourcedString.concat([s.upper() for s in self.substrings])

    def swapcase(self):
        return SourcedString.concat([s.swapcase() for s in self.substrings])

    def title(self):
        return SourcedString.concat([s.title() for s in self.substrings])

    def encode(self, encoding=None, errors='strict'):
        return SourcedString.concat([s.encode(encoding, errors)
                                     for s in self.substrings])

    def _decode_one_to_one(self, unicode_chars):
        index = 0
        result = []
        for substring in self.substrings:
            decoded_substring = unicode_chars[index:index+len(substring)]
            if isinstance(substring, SourcedString):
                result.append(SourcedString(decoded_substring, substring.source))
            else:
                result.append(decoded_substring)
            index += len(substring)
        return SourcedString.concat(result)

#//////////////////////////////////////////////////////////////////////
# Concrete Sourced String Classes
#//////////////////////////////////////////////////////////////////////

class SimpleSourcedByteString(SimpleSourcedString, str):
    _stringtype = str
class SimpleSourcedUnicodeString(SimpleSourcedString, unicode):
    _stringtype = unicode
class CompoundSourcedByteString(CompoundSourcedString, str):
    _stringtype = str
class CompoundSourcedUnicodeString(CompoundSourcedString, unicode):
    _stringtype = unicode
    def __init__(self, substrings):
        # If any substrings have type 'str', then decode them to unicode.
        for i in range(len(substrings)):
            if not isinstance(substrings[i], unicode):
                substrings[i] = substrings[i].decode()
        CompoundSourcedString.__init__(self, substrings)

#//////////////////////////////////////////////////////////////////////
# Sourced String Regexp
#//////////////////////////////////////////////////////////////////////

_original_re_compile = re.compile
_original_re_sub = re.sub
_original_re_subn = re.subn

class SourcedStringRegexp(object):
    """
    Wrapper for regexp pattern objects that cause the ``sub`` and
    ``subn`` methods to return sourced strings.
    """
    def __init__(self, pattern, flags=0):
        if isinstance(pattern, basestring):
            pattern = _original_re_compile(pattern, flags)
        self.pattern = pattern
    def __getattr__(self, attr):
        return getattr(self.pattern, attr)

    def subn(self, repl, string, count=0):
        if (isinstance(repl, SourcedString) or
            isinstance(string, SourcedString)):
            result = ''
            pos = 0
            n = 0
            for match in self.pattern.finditer(string):
                result += string[pos:match.start()]
                result += repl
                pos = match.end()
                n += 1
                if count and n==count: break
            result += string[pos:]
            return result, n
        else:
            return self.pattern.subn(repl, string, count)

    def sub(self, repl, string, count=0):
        return self.subn(repl, string, count)[0]

    @staticmethod
    def patch_re_module():
        """
        Modify the standard ``re`` module by installing new versions of
        the functions ``re.compile``, ``re.sub``, and ``re.subn``,
        causing regular expression substitutions to return
        ``SourcedStrings`` when called with ``SourcedStrings``
        arguments.

        Use this function only if necessary: it potentially affects
        all Python modules that use regular expressions!
        """
        def new_re_sub(pattern, repl, string, count=0):
            return re.compile(pattern).sub(repl, string, count)
        def new_re_subn(pattern, repl, string, count=0):
            return re.compile(pattern).subn(repl, string, count)
        re.compile = SourcedStringRegexp
        re.sub = new_re_sub
        re.subn = new_re_subn

    @staticmethod
    def unpatch_re_module():
        """
        Restore the standard ``re`` module to its original state
        (undoing the work that was done by ``patch_re_module()``).
        """
        re.compile = _original_re_compile
        re.sub = _original_re_sub
        re.subn = _original_re_subn


#//////////////////////////////////////////////////////////////////////
# Sourced String Stream
#//////////////////////////////////////////////////////////////////////

class SourcedStringStream(object):
    """
    Wrapper for a read-only stream that causes ``read()`` (and related
    methods) to return ``SourcedStringBase``.
    ``seek()`` and ``tell()`` are supported, but (currently) there are
    some restrictions on the values that may be passed to ``seek()``.
    """
    def __init__(self, stream, docid=None, byte_offsets=False):
        self.stream = stream
        """The underlying stream."""

        self.docid = docid
        """The docid attribute for sourced strings"""

        self.charpos = 0
        """The current character (not byte) position"""

        assert not byte_offsets, 'not supported yet!'

    #/////////////////////////////////////////////////////////////////
    # Read methods
    #/////////////////////////////////////////////////////////////////

    def read(self, size=None):
        if size is None: return self._sourced_string(self.stream.read())
        else: return self._sourced_string(self.stream.read(size))

    def readline(self, size=None):
        if size is None:  return self._sourced_string(self.stream.readline())
        else: return self._sourced_string(self.stream.readline(size))

    def readlines(self, sizehint=None, keepends=True):
        """
        Read this file's contents, decode them using this reader's
        encoding, and return it as a list of unicode lines.

        :rtype: list(unicode)
        :param sizehint: Ignored.
        :param keepends: If false, then strip newlines.
        """
        return self.read().splitlines(keepends)

    def next(self):
        """Return the next decoded line from the underlying stream."""
        line = self.readline()
        if line: return line
        else: raise StopIteration

    def __iter__(self):
        """Return self"""
        return self

    def xreadlines(self):
        """Return self"""
        return self

    def _sourced_string(self, contents):
        """Turn the given string into an sourced string, and update
           charpos."""
        # [xx] currently we only support character offsets, not byte
        # offsets!
        source = ConsecutiveCharStringSource(self.docid, self.charpos,
                                             self.charpos+len(contents))
        self.charpos += len(contents)
        return SourcedString(contents, source)

    #/////////////////////////////////////////////////////////////////
    # Pass-through methods & properties
    #/////////////////////////////////////////////////////////////////

    @property
    def closed(self): 
        """True if the underlying stream is closed."""
        return self.stream.closed

    @property
    def name(self): 
        """The name of the underlying stream."""
        return self.stream.name

    @property
    def mode(self): 
        """The mode of the underlying stream."""
        return self.stream.mode

    def close(self):
        """Close the underlying stream."""
        self.stream.close()

    #/////////////////////////////////////////////////////////////////
    # Seek and tell
    #/////////////////////////////////////////////////////////////////

    class SourcedStringStreamPos(int):
        def __new__(cls, bytepos, charpos):
            self = int.__new__(cls, bytepos)
            self.charpos = charpos
            return self

    def seek(self, offset, whence=0):
        if whence == 0:
            if isinstance(offset, self.SourcedStringStreamPos):
                self.stream.seek(offset)
                self.charpos = offset.charpos
            elif offset == 0:
                self.stream.seek(0)
                self.charpos = 0
            else:
                raise TypeError('seek() must be called with a value that '
                                'was returned by tell().')
        elif whence == 1:
            raise TypeError('Relative seek not supported for '
                            'SourcedStringStream.')
        elif whence == 2:
            raise TypeError('Seek-from-end not supported for '
                            'SourcedStringStream.')
        else:
            raise ValueError('Bad whence value %r' % whence)

    def tell(self):
        bytepos = self.stream.tell()
        return self.SourcedStringStreamPos(bytepos, self.charpos)






# Natural Language Toolkit: API for Corpus Readers
#
# Copyright (C) 2001-2012 NLTK Project
# Author: Steven Bird <sb@ldc.upenn.edu>
#         Edward Loper <edloper@gradient.cis.upenn.edu>
# URL: <http://www.nltk.org/>
# For license information, see LICENSE.TXT

"""
API for corpus readers.
"""

import os
import re
from collections import defaultdict

from nltk.data import PathPointer, FileSystemPathPointer, ZipFilePathPointer
from nltk.sourcedstring import SourcedStringStream

from util import *

class CorpusReader(object):
    """
    A base class for "corpus reader" classes, each of which can be
    used to read a specific corpus format.  Each individual corpus
    reader instance is used to read a specific corpus, consisting of
    one or more files under a common root directory.  Each file is
    identified by its ``file identifier``, which is the relative path
    to the file from the root directory.

    A separate subclass is be defined for each corpus format.  These
    subclasses define one or more methods that provide 'views' on the
    corpus contents, such as ``words()`` (for a list of words) and
    ``parsed_sents()`` (for a list of parsed sentences).  Called with
    no arguments, these methods will return the contents of the entire
    corpus.  For most corpora, these methods define one or more
    selection arguments, such as ``fileids`` or ``categories``, which can
    be used to select which portion of the corpus should be returned.
    """
    def __init__(self, root, fileids, encoding=None, tag_mapping_function=None):
        """
        :type root: PathPointer or str
        :param root: A path pointer identifying the root directory for
            this corpus.  If a string is specified, then it will be
            converted to a ``PathPointer`` automatically.
        :param fileids: A list of the files that make up this corpus.
            This list can either be specified explicitly, as a list of
            strings; or implicitly, as a regular expression over file
            paths.  The absolute path for each file will be constructed
            by joining the reader's root to each file name.
        :param encoding: The default unicode encoding for the files
            that make up the corpus.  The value of ``encoding`` can be any
            of the following:
            - A string: ``encoding`` is the encoding name for all files.
            - A dictionary: ``encoding[file_id]`` is the encoding
              name for the file whose identifier is ``file_id``.  If
              ``file_id`` is not in ``encoding``, then the file
              contents will be processed using non-unicode byte strings.
            - A list: ``encoding`` should be a list of ``(regexp, encoding)``
              tuples.  The encoding for a file whose identifier is ``file_id``
              will be the ``encoding`` value for the first tuple whose
              ``regexp`` matches the ``file_id``.  If no tuple's ``regexp``
              matches the ``file_id``, the file contents will be processed
              using non-unicode byte strings.
            - None: the file contents of all files will be
              processed using non-unicode byte strings.
        :param tag_mapping_function: A function for normalizing or
                simplifying the POS tags returned by the tagged_words()
                or tagged_sents() methods.
        """
        # Convert the root to a path pointer, if necessary.
        if isinstance(root, basestring):
            m = re.match('(.*\.zip)/?(.*)$|', root)
            zipfile, zipentry = m.groups()
            if zipfile:
                root = ZipFilePathPointer(zipfile, zipentry)
            else:
                root = FileSystemPathPointer(root)
        elif not isinstance(root, PathPointer):
            raise TypeError('CorpusReader: expected a string or a PathPointer')

        # If `fileids` is a regexp, then expand it.
        if isinstance(fileids, basestring):
            fileids = find_corpus_fileids(root, fileids)

        self._fileids = fileids
        """A list of the relative paths for the fileids that make up
        this corpus."""

        self._root = root
        """The root directory for this corpus."""

        # If encoding was specified as a list of regexps, then convert
        # it to a dictionary.
        if isinstance(encoding, list):
            encoding_dict = {}
            for fileid in self._fileids:
                for x in encoding:
                    (regexp, enc) = x
                    if re.match(regexp, fileid):
                        encoding_dict[fileid] = enc
                        break
            encoding = encoding_dict

        self._encoding = encoding
        """The default unicode encoding for the fileids that make up
           this corpus.  If ``encoding`` is None, then the file
           contents are processed using byte strings (str)."""
        self._tag_mapping_function = tag_mapping_function

    def __repr__(self):
        if isinstance(self._root, ZipFilePathPointer):
            path = '%s/%s' % (self._root.zipfile.filename, self._root.entry)
        else:
            path = '%s' % self._root.path
        return '<%s in %r>' % (self.__class__.__name__, path)

    def readme(self):
        """
        Return the contents of the corpus README file, if it exists.
        """

        return self.open("README").read()

    def fileids(self):
        """
        Return a list of file identifiers for the fileids that make up
        this corpus.
        """
        return self._fileids

    def abspath(self, fileid):
        """
        Return the absolute path for the given file.

        :type file: str
        :param file: The file identifier for the file whose path
            should be returned.
        :rtype: PathPointer
        """
        return self._root.join(fileid)

    def abspaths(self, fileids=None, include_encoding=False,
                 include_fileid=False):
        """
        Return a list of the absolute paths for all fileids in this corpus;
        or for the given list of fileids, if specified.

        :type fileids: None or str or list
        :param fileids: Specifies the set of fileids for which paths should
            be returned.  Can be None, for all fileids; a list of
            file identifiers, for a specified set of fileids; or a single
            file identifier, for a single file.  Note that the return
            value is always a list of paths, even if ``fileids`` is a
            single file identifier.

        :param include_encoding: If true, then return a list of
            ``(path_pointer, encoding)`` tuples.

        :rtype: list(PathPointer)
        """
        if fileids is None:
            fileids = self._fileids
        elif isinstance(fileids, basestring):
            fileids = [fileids]

        paths = [self._root.join(f) for f in fileids]

        if include_encoding and include_fileid:
            return zip(paths, [self.encoding(f) for f in fileids], fileids)
        elif include_fileid:
            return zip(paths, fileids)
        elif include_encoding:
            return zip(paths, [self.encoding(f) for f in fileids])
        else:
            return paths

    def open(self, file, sourced=False):
        """
        Return an open stream that can be used to read the given file.
        If the file's encoding is not None, then the stream will
        automatically decode the file's contents into unicode.

        :param file: The file identifier of the file to read.
        """
        encoding = self.encoding(file)
        stream = self._root.join(file).open(encoding)
        if sourced:
            stream = SourcedStringStream(stream, file)
        return stream

    def encoding(self, file):
        """
        Return the unicode encoding for the given corpus file, if known.
        If the encoding is unknown, or if the given file should be
        processed using byte strings (str), then return None.
        """
        if isinstance(self._encoding, dict):
            return self._encoding.get(file)
        else:
            return self._encoding

    def _get_root(self): return self._root
    root = property(_get_root, doc="""
        The directory where this corpus is stored.

        :type: PathPointer""")


######################################################################
#{ Corpora containing categorized items
######################################################################

class CategorizedCorpusReader(object):
    """
    A mixin class used to aid in the implementation of corpus readers
    for categorized corpora.  This class defines the method
    ``categories()``, which returns a list of the categories for the
    corpus or for a specified set of fileids; and overrides ``fileids()``
    to take a ``categories`` argument, restricting the set of fileids to
    be returned.

    Subclasses are expected to:

      - Call ``__init__()`` to set up the mapping.

      - Override all view methods to accept a ``categories`` parameter,
        which can be used *instead* of the ``fileids`` parameter, to
        select which fileids should be included in the returned view.
    """

    def __init__(self, kwargs):
        """
        Initialize this mapping based on keyword arguments, as
        follows:

          - cat_pattern: A regular expression pattern used to find the
            category for each file identifier.  The pattern will be
            applied to each file identifier, and the first matching
            group will be used as the category label for that file.

          - cat_map: A dictionary, mapping from file identifiers to
            category labels.

          - cat_file: The name of a file that contains the mapping
            from file identifiers to categories.  The argument
            ``cat_delimiter`` can be used to specify a delimiter.

        The corresponding argument will be deleted from ``kwargs``.  If
        more than one argument is specified, an exception will be
        raised.
        """
        self._f2c = None #: file-to-category mapping
        self._c2f = None #: category-to-file mapping

        self._pattern = None #: regexp specifying the mapping
        self._map = None #: dict specifying the mapping
        self._file = None #: fileid of file containing the mapping
        self._delimiter = None #: delimiter for ``self._file``

        if 'cat_pattern' in kwargs:
            self._pattern = kwargs['cat_pattern']
            del kwargs['cat_pattern']
        elif 'cat_map' in kwargs:
            self._map = kwargs['cat_map']
            del kwargs['cat_map']
        elif 'cat_file' in kwargs:
            self._file = kwargs['cat_file']
            del kwargs['cat_file']
            if 'cat_delimiter' in kwargs:
                self._delimiter = kwargs['cat_delimiter']
                del kwargs['cat_delimiter']
        else:
            raise ValueError('Expected keyword argument cat_pattern or '
                             'cat_map or cat_file.')


        if ('cat_pattern' in kwargs or 'cat_map' in kwargs or
            'cat_file' in kwargs):
            raise ValueError('Specify exactly one of: cat_pattern, '
                             'cat_map, cat_file.')

    def _init(self):
        self._f2c = defaultdict(set)
        self._c2f = defaultdict(set)

        if self._pattern is not None:
            for file_id in self._fileids:
                category = re.match(self._pattern, file_id).group(1)
                self._add(file_id, category)

        elif self._map is not None:
            for (file_id, categories) in self._map.items():
                for category in categories:
                    self._add(file_id, category)

        elif self._file is not None:
            for line in self.open(self._file).readlines():
                line = line.strip()
                file_id, categories = line.split(self._delimiter, 1)
                if file_id not in self.fileids():
                    raise ValueError('In category mapping file %s: %s '
                                     'not found' % (self._file, file_id))
                for category in categories.split(self._delimiter):
                    self._add(file_id, category)

    def _add(self, file_id, category):
        self._f2c[file_id].add(category)
        self._c2f[category].add(file_id)

    def categories(self, fileids=None):
        """
        Return a list of the categories that are defined for this corpus,
        or for the file(s) if it is given.
        """
        if self._f2c is None:
            self._init()
        if fileids is None:
            return sorted(self._c2f)
        if isinstance(fileids, basestring):
            fileids = [fileids]
        return sorted(set.union(*[self._f2c[d] for d in fileids]))

    def fileids(self, categories=None):
        """
        Return a list of file identifiers for the files that make up
        this corpus, or that make up the given category(s) if specified.
        """
        if categories is None:
            return super(CategorizedCorpusReader, self).fileids()
        elif isinstance(categories, basestring):
            if self._f2c is None:
                self._init()
            if categories in self._c2f:
                return sorted(self._c2f[categories])
            else:
                raise ValueError('Category %s not found' % categories)
        else:
            if self._f2c is None:
                self._init()
            return sorted(set.union(*[self._c2f[c] for c in categories]))

######################################################################
#{ Treebank readers
######################################################################

#[xx] is it worth it to factor this out?
class SyntaxCorpusReader(CorpusReader):
    """
    An abstract base class for reading corpora consisting of
    syntactically parsed text.  Subclasses should define:

      - ``__init__``, which specifies the location of the corpus
        and a method for detecting the sentence blocks in corpus files.
      - ``_read_block``, which reads a block from the input stream.
      - ``_word``, which takes a block and returns a list of list of words.
      - ``_tag``, which takes a block and returns a list of list of tagged
        words.
      - ``_parse``, which takes a block and returns a list of parsed
        sentences.
    """
    def _parse(self, s):
        raise NotImplementedError()
    def _word(self, s):
        raise NotImplementedError()
    def _tag(self, s):
        raise NotImplementedError()
    def _read_block(self, stream):
        raise NotImplementedError()

    def raw(self, fileids=None):
        if fileids is None: fileids = self._fileids
        elif isinstance(fileids, basestring): fileids = [fileids]
        return concat([self.open(f).read() for f in fileids])

    def parsed_sents(self, fileids=None):
        reader = self._read_parsed_sent_block
        return concat([StreamBackedCorpusView(fileid, reader, encoding=enc)
                       for fileid, enc in self.abspaths(fileids, True)])

    def tagged_sents(self, fileids=None, simplify_tags=False):
        def reader(stream):
            return self._read_tagged_sent_block(stream, simplify_tags)
        return concat([StreamBackedCorpusView(fileid, reader, encoding=enc)
                       for fileid, enc in self.abspaths(fileids, True)])

    def sents(self, fileids=None):
        reader = self._read_sent_block
        return concat([StreamBackedCorpusView(fileid, reader, encoding=enc)
                       for fileid, enc in self.abspaths(fileids, True)])

    def tagged_words(self, fileids=None, simplify_tags=False):
        def reader(stream):
            return self._read_tagged_word_block(stream, simplify_tags)
        return concat([StreamBackedCorpusView(fileid, reader, encoding=enc)
                       for fileid, enc in self.abspaths(fileids, True)])

    def words(self, fileids=None):
        return concat([StreamBackedCorpusView(fileid,
                                              self._read_word_block,
                                              encoding=enc)
                       for fileid, enc in self.abspaths(fileids, True)])

    #------------------------------------------------------------
    #{ Block Readers

    def _read_word_block(self, stream):
        return sum(self._read_sent_block(stream), [])

    def _read_tagged_word_block(self, stream, simplify_tags=False):
        return sum(self._read_tagged_sent_block(stream, simplify_tags), [])

    def _read_sent_block(self, stream):
        return filter(None, [self._word(t) for t in self._read_block(stream)])

    def _read_tagged_sent_block(self, stream, simplify_tags=False):
        return filter(None, [self._tag(t, simplify_tags)
                             for t in self._read_block(stream)])

    def _read_parsed_sent_block(self, stream):
        return filter(None, [self._parse(t) for t in self._read_block(stream)])

    #} End of Block Readers
    #------------------------------------------------------------




# Natural Language Toolkit: Models for first-order languages with lambda
#
# Copyright (C) 2001-2012 NLTK Project
# Author: Ewan Klein <ewan@inf.ed.ac.uk>,
# URL: <http://nltk.sourceforge.net>
# For license information, see LICENSE.TXT

#TODO:
    #- fix tracing
    #- fix iterator-based approach to existentials

"""
This module provides data structures for representing first-order
models.
"""

from pprint import pformat
import inspect
import textwrap

from nltk.decorators import decorator

from nltk.sem.logic import (AbstractVariableExpression, AllExpression,
                            AndExpression, ApplicationExpression, EqualityExpression,
                            ExistsExpression, IffExpression, ImpExpression,
                            IndividualVariableExpression, LambdaExpression,
                            LogicParser, NegatedExpression, OrExpression,
                            Variable, is_indvar)


class Error(Exception): pass

class Undefined(Error):  pass

def trace(f, *args, **kw):
    argspec = inspect.getargspec(f)
    d = dict(zip(argspec[0], args))
    if d.pop('trace', None):
        print
        for item in d.items():
            print "%s => %s" % item
    return f(*args, **kw)

def is_rel(s):
    """
    Check whether a set represents a relation (of any arity).

    :param s: a set containing tuples of str elements
    :type s: set
    :rtype: bool
        """
    # we have the empty relation, i.e. set()
    if len(s) == 0:
        return True
    # all the elements are tuples of the same length
    elif s == set([elem for elem in s if isinstance(elem, tuple)]) and\
         len(max(s))==len(min(s)):
        return True
    else:
        raise ValueError, "Set %r contains sequences of different lengths" % s

def set2rel(s):
    """
    Convert a set containing individuals (strings or numbers) into a set of
    unary tuples. Any tuples of strings already in the set are passed through
    unchanged.

    For example:
      - set(['a', 'b']) => set([('a',), ('b',)])
      - set([3, 27]) => set([('3',), ('27',)])

    :type s: set
    :rtype: set of tuple of str
    """
    new = set()
    for elem in s:
        if isinstance(elem, str):
            new.add((elem,))
        elif isinstance(elem, int):
            new.add((str(elem,)))
        else:
            new.add(elem)
    return new

def arity(rel):
    """
    Check the arity of a relation.
    :type rel: set of tuples
    :rtype: int of tuple of str
    """
    if len(rel) == 0:
        return 0
    return len(list(rel)[0])


class Valuation(dict):
    """
    A dictionary which represents a model-theoretic Valuation of non-logical constants.
    Keys are strings representing the constants to be interpreted, and values correspond
    to individuals (represented as strings) and n-ary relations (represented as sets of tuples
    of strings).

    An instance of ``Valuation`` will raise a KeyError exception (i.e.,
    just behave like a standard  dictionary) if indexed with an expression that
    is not in its list of symbols.
    """
    def __init__(self, iter):
        """
        :param iter: a list of (symbol, value) pairs.
        """
        dict.__init__(self)
        for (sym, val) in iter:
            if isinstance(val, str) or isinstance(val, bool):
                self[sym] = val
            elif isinstance(val, set):
                self[sym] = set2rel(val)
            else:
                msg = textwrap.fill("Error in initializing Valuation. "
                                    "Unrecognized value for symbol '%s':\n%s" % (sym, val), width=66)

                raise ValueError(msg)

    def __getitem__(self, key):
        if key in self:
            return dict.__getitem__(self, key)
        else:
            raise Undefined,  "Unknown expression: '%s'" % key

    def __str__(self):
        return pformat(self)

    @property
    def domain(self):
        """Set-theoretic domain of the value-space of a Valuation."""
        dom = []
        for val in self.values():
            if isinstance(val, str):
                dom.append(val)
            elif not isinstance(val, bool):
                dom.extend([elem for tuple in val for elem in tuple if elem is not None])
        return set(dom)

    @property
    def symbols(self):
        """The non-logical constants which the Valuation recognizes."""
        return sorted(self.keys())


class Assignment(dict):
    """
    A dictionary which represents an assignment of values to variables.

    An assigment can only assign values from its domain.

    If an unknown expression *a* is passed to a model *M*\ 's
    interpretation function *i*, *i* will first check whether *M*\ 's
    valuation assigns an interpretation to *a* as a constant, and if
    this fails, *i* will delegate the interpretation of *a* to
    *g*. *g* only assigns values to individual variables (i.e.,
    members of the class ``IndividualVariableExpression`` in the ``logic``
    module. If a variable is not assigned a value by *g*, it will raise
    an ``Undefined`` exception.

    A variable *Assignment* is a mapping from individual variables to
    entities in the domain. Individual variables are usually indicated
    with the letters ``'x'``, ``'y'``, ``'w'`` and ``'z'``, optionally
    followed by an integer (e.g., ``'x0'``, ``'y332'``).  Assignments are
    created using the ``Assignment`` constructor, which also takes the
    domain as a parameter.

        >>> from nltk.sem.evaluate import Assignment
        >>> dom = set(['u1', 'u2', 'u3', 'u4'])
        >>> g3 = Assignment(dom, [('x', 'u1'), ('y', 'u2')])
        >>> g3
        {'y': 'u2', 'x': 'u1'}

    There is also a ``print`` format for assignments which uses a notation
    closer to that in logic textbooks:

        >>> print g3
        g[u2/y][u1/x]

    It is also possible to update an assignment using the ``add`` method:

        >>> dom = set(['u1', 'u2', 'u3', 'u4'])
        >>> g4 = Assignment(dom)
        >>> g4.add('x', 'u1')
        {'x': 'u1'}

    With no arguments, ``purge()`` is equivalent to ``clear()`` on a dictionary:

        >>> g4.purge()
        >>> g4
        {}

    :param domain: the domain of discourse
    :type domain: set
    :param assign: a list of (varname, value) associations
    :type assign: list
    """

    def __init__(self, domain, assign=None):
        dict.__init__(self)
        self.domain = domain
        if assign:
            for (var, val) in assign:
                assert val in self.domain,\
                       "'%s' is not in the domain: %s" % (val, self.domain)
                assert is_indvar(var),\
                       "Wrong format for an Individual Variable: '%s'" % var
                self[var] = val
        self._addvariant()

    def __getitem__(self, key):
        if key in self:
            return dict.__getitem__(self, key)
        else:
            raise Undefined, "Not recognized as a variable: '%s'" % key

    def copy(self):
        new = Assignment(self.domain)
        new.update(self)
        return new

    def purge(self, var=None):
        """
        Remove one or all keys (i.e. logic variables) from an
        assignment, and update ``self.variant``.

        :param var: a Variable acting as a key for the assignment.
        """
        if var:
            val = self[var]
            del self[var]
        else:
            self.clear()
        self._addvariant()
        return None

    def __str__(self):
        """
        Pretty printing for assignments. {'x', 'u'} appears as 'g[u/x]'
        """
        gstring = "g"
        for (val, var) in self.variant:
            gstring += "[%s/%s]" % (val, var)
        return gstring

    def _addvariant(self):
        """
        Create a more pretty-printable version of the assignment.
        """
        list = []
        for item in self.items():
            pair = (item[1], item[0])
            list.append(pair)
        self.variant = list
        return None

    def add(self, var, val):
        """
        Add a new variable-value pair to the assignment, and update
        ``self.variant``.

        """
        assert val in self.domain,\
               "%s is not in the domain %s" % (val, self.domain)
        assert is_indvar(var),\
               "Wrong format for an Individual Variable: '%s'" % var
        self[var] = val
        self._addvariant()
        return self


class Model(object):
    """
    A first order model is a domain *D* of discourse and a valuation *V*.

    A domain *D* is a set, and a valuation *V* is a map that associates
    expressions with values in the model.
    The domain of *V* should be a subset of *D*.

    Construct a new ``Model``.

    :type domain: set
    :param domain: A set of entities representing the domain of discourse of the model.
    :type valuation: Valuation
    :param valuation: the valuation of the model.
    :param prop: If this is set, then we are building a propositional\
    model and don't require the domain of *V* to be subset of *D*.
    """

    def __init__(self, domain, valuation):
        assert isinstance(domain, set)
        self.domain = domain
        self.valuation = valuation
        if not domain.issuperset(valuation.domain):
            raise Error,\
                  "The valuation domain, %s, must be a subset of the model's domain, %s"\
                  % (valuation.domain, domain)

    def __repr__(self):
        return "(%r, %r)" % (self.domain, self.valuation)

    def __str__(self):
        return "Domain = %s,\nValuation = \n%s" % (self.domain, self.valuation)

    def evaluate(self, expr, g, trace=None):
        """
        Call the ``LogicParser`` to parse input expressions, and
        provide a handler for ``satisfy``
        that blocks further propagation of the ``Undefined`` error.
        :param expr: An ``Expression`` of ``logic``.
        :type g: Assignment
        :param g: an assignment to individual variables.
        :rtype: bool or 'Undefined'
        """
        try:
            lp = LogicParser()
            parsed = lp.parse(expr)
            value = self.satisfy(parsed, g, trace=trace)
            if trace:
                print
                print "'%s' evaluates to %s under M, %s" %  (expr, value, g)
            return value
        except Undefined:
            if trace:
                print
                print "'%s' is undefined under M, %s" %  (expr, g)
            return 'Undefined'


    def satisfy(self, parsed, g, trace=None):
        """
        Recursive interpretation function for a formula of first-order logic.

        Raises an ``Undefined`` error when ``parsed`` is an atomic string
        but is not a symbol or an individual variable.

        :return: Returns a truth value or ``Undefined`` if ``parsed`` is\
        complex, and calls the interpretation function ``i`` if ``parsed``\
        is atomic.

        :param parsed: An expression of ``logic``.
        :type g: Assignment
        :param g: an assignment to individual variables.
        """

        if isinstance(parsed, ApplicationExpression):
            function, arguments = parsed.uncurry()
            if isinstance(function, AbstractVariableExpression):
                #It's a predicate expression ("P(x,y)"), so used uncurried arguments
                funval = self.satisfy(function, g)
                argvals = tuple([self.satisfy(arg, g) for arg in arguments])
                return argvals in funval
            else:
                #It must be a lambda expression, so use curried form
                funval = self.satisfy(parsed.function, g)
                argval = self.satisfy(parsed.argument, g)
                return funval[argval]
        elif isinstance(parsed, NegatedExpression):
            return not self.satisfy(parsed.term, g)
        elif isinstance(parsed, AndExpression):
            return self.satisfy(parsed.first, g) and \
                   self.satisfy(parsed.second, g)
        elif isinstance(parsed, OrExpression):
            return self.satisfy(parsed.first, g) or \
                   self.satisfy(parsed.second, g)
        elif isinstance(parsed, ImpExpression):
            return (not self.satisfy(parsed.first, g)) or \
                   self.satisfy(parsed.second, g)
        elif isinstance(parsed, IffExpression):
            return self.satisfy(parsed.first, g) == \
                   self.satisfy(parsed.second, g)
        elif isinstance(parsed, EqualityExpression):
            return self.satisfy(parsed.first, g) == \
                   self.satisfy(parsed.second, g)
        elif isinstance(parsed, AllExpression):
            new_g = g.copy()
            for u in self.domain:
                new_g.add(parsed.variable.name, u)
                if not self.satisfy(parsed.term, new_g):
                    return False
            return True
        elif isinstance(parsed, ExistsExpression):
            new_g = g.copy()
            for u in self.domain:
                new_g.add(parsed.variable.name, u)
                if self.satisfy(parsed.term, new_g):
                    return True
            return False
        elif isinstance(parsed, LambdaExpression):
            cf = {}
            var = parsed.variable.name
            for u in self.domain:
                val = self.satisfy(parsed.term, g.add(var, u))
                # NB the dict would be a lot smaller if we do this:
                # if val: cf[u] = val
                # But then need to deal with cases where f(a) should yield
                # a function rather than just False.
                cf[u] = val
            return cf
        else:
            return self.i(parsed, g, trace)

    #@decorator(trace_eval)
    def i(self, parsed, g, trace=False):
        """
        An interpretation function.

        Assuming that ``parsed`` is atomic:

        - if ``parsed`` is a non-logical constant, calls the valuation *V*
        - else if ``parsed`` is an individual variable, calls assignment *g*
        - else returns ``Undefined``.

        :param parsed: an ``Expression`` of ``logic``.
        :type g: Assignment
        :param g: an assignment to individual variables.
        :return: a semantic value
        """
        # If parsed is a propositional letter 'p', 'q', etc, it could be in valuation.symbols
        # and also be an IndividualVariableExpression. We want to catch this first case.
        # So there is a procedural consequence to the ordering of clauses here:
        if parsed.variable.name in self.valuation.symbols:
            return self.valuation[parsed.variable.name]
        elif isinstance(parsed, IndividualVariableExpression):
            return g[parsed.variable.name]

        else:
            raise Undefined, "Can't find a value for %s" % parsed

    def satisfiers(self, parsed, varex, g, trace=None, nesting=0):
        """
        Generate the entities from the model's domain that satisfy an open formula.

        :param parsed: an open formula
        :type parsed: Expression
        :param varex: the relevant free individual variable in ``parsed``.
        :type varex: VariableExpression or str
        :param g: a variable assignment
        :type g:  Assignment
        :return: a set of the entities that satisfy ``parsed``.
        """

        spacer = '   '
        indent = spacer + (spacer * nesting)
        candidates = []

        if isinstance(varex, str):
            var = Variable(varex)
        else:
            var = varex

        if var in parsed.free():
            if trace:
                print
                print (spacer * nesting) + "Open formula is '%s' with assignment %s" % (parsed, g)
            for u in self.domain:
                new_g = g.copy()
                new_g.add(var.name, u)
                if trace > 1:
                    lowtrace = trace-1
                else:
                    lowtrace = 0
                value = self.satisfy(parsed, new_g, lowtrace)

                if trace:
                    print indent + "(trying assignment %s)" % new_g

                # parsed == False under g[u/var]?
                if value == False:
                    if trace:
                        print  indent + "value of '%s' under %s is False" % (parsed, new_g)

                # so g[u/var] is a satisfying assignment
                else:
                    candidates.append(u)
                    if trace:
                        print indent + "value of '%s' under %s is %s" % (parsed, new_g, value)

            result = set(c for c in candidates)
        # var isn't free in parsed
        else:
            raise Undefined, "%s is not free in %s" % (var.name, parsed)

        return result





#//////////////////////////////////////////////////////////////////////
# Demo..
#//////////////////////////////////////////////////////////////////////
# number of spacer chars
mult = 30

# Demo 1: Propositional Logic
#################
def propdemo(trace=None):
    """Example of a propositional model."""

    global val1, dom1, m1, g1
    val1 = Valuation([('P', True), ('Q', True), ('R', False)])
    dom1 = set([])
    m1 = Model(dom1, val1)
    g1 = Assignment(dom1)

    print
    print '*' * mult
    print "Propositional Formulas Demo"
    print '*' * mult
    print '(Propositional constants treated as nullary predicates)'
    print
    print "Model m1:\n", m1
    print '*' * mult
    sentences = [
    '(P & Q)',
    '(P & R)',
    '- P',
    '- R',
    '- - P',
    '- (P & R)',
    '(P | R)',
    '(R | P)',
    '(R | R)',
    '(- P | R)',
    '(P | - P)',
    '(P -> Q)',
    '(P -> R)',
    '(R -> P)',
    '(P <-> P)',
    '(R <-> R)',
    '(P <-> R)',
    ]

    for sent in sentences:
        if trace:
            print
            m1.evaluate(sent, g1, trace)
        else:
            print "The value of '%s' is: %s" % (sent, m1.evaluate(sent, g1))

# Demo 2: FOL Model
#############

def folmodel(quiet=False, trace=None):
    """Example of a first-order model."""

    global val2, v2, dom2, m2, g2

    v2 = [('adam', 'b1'), ('betty', 'g1'), ('fido', 'd1'),\
         ('girl', set(['g1', 'g2'])), ('boy', set(['b1', 'b2'])), ('dog', set(['d1'])),
         ('love', set([('b1', 'g1'), ('b2', 'g2'), ('g1', 'b1'), ('g2', 'b1')]))]
    val2 = Valuation(v2)
    dom2 = val2.domain
    m2 = Model(dom2, val2)
    g2 = Assignment(dom2, [('x', 'b1'), ('y', 'g2')])

    if not quiet:
        print
        print '*' * mult
        print "Models Demo"
        print "*" * mult
        print "Model m2:\n", "-" * 14,"\n", m2
        print "Variable assignment = ", g2

        exprs = ['adam', 'boy', 'love', 'walks', 'x', 'y', 'z']
        lp = LogicParser()
        parsed_exprs = [lp.parse(e) for e in exprs]

        print
        for parsed in parsed_exprs:
            try:
                print "The interpretation of '%s' in m2 is %s" % (parsed, m2.i(parsed, g2))
            except Undefined:
                print "The interpretation of '%s' in m2 is Undefined" % parsed


        applications = [('boy', ('adam')), ('walks', ('adam',)), ('love', ('adam', 'y')), ('love', ('y', 'adam'))]

        for (fun, args) in applications:
            try:
                funval = m2.i(lp.parse(fun), g2)
                argsval = tuple(m2.i(lp.parse(arg), g2) for arg in args)
                print "%s(%s) evaluates to %s" % (fun, args, argsval in funval)
            except Undefined:
                print "%s(%s) evaluates to Undefined" % (fun, args)

# Demo 3: FOL
#########

def foldemo(trace=None):
    """
    Interpretation of closed expressions in a first-order model.
    """
    folmodel(quiet=True)

    print
    print '*' * mult
    print "FOL Formulas Demo"
    print '*' * mult

    formulas = [
    'love (adam, betty)',
    '(adam = mia)',
    '\\x. (boy(x) | girl(x))',
    '\\x. boy(x)(adam)',
    '\\x y. love(x, y)',
    '\\x y. love(x, y)(adam)(betty)',
    '\\x y. love(x, y)(adam, betty)',
    '\\x y. (boy(x) & love(x, y))',
    '\\x. exists y. (boy(x) & love(x, y))',
    'exists z1. boy(z1)',
    'exists x. (boy(x) &  -(x = adam))',
    'exists x. (boy(x) & all y. love(y, x))',
    'all x. (boy(x) | girl(x))',
    'all x. (girl(x) -> exists y. boy(y) & love(x, y))',    #Every girl loves exists boy.
    'exists x. (boy(x) & all y. (girl(y) -> love(y, x)))',  #There is exists boy that every girl loves.
    'exists x. (boy(x) & all y. (girl(y) -> love(x, y)))',  #exists boy loves every girl.
    'all x. (dog(x) -> - girl(x))',
    'exists x. exists y. (love(x, y) & love(x, y))'
    ]


    for fmla in formulas:
        g2.purge()
        if trace:
            m2.evaluate(fmla, g2, trace)
        else:
            print "The value of '%s' is: %s" % (fmla, m2.evaluate(fmla, g2))


# Demo 3: Satisfaction
#############

def satdemo(trace=None):
    """Satisfiers of an open formula in a first order model."""

    print
    print '*' * mult
    print "Satisfiers Demo"
    print '*' * mult

    folmodel(quiet=True)

    formulas = [
               'boy(x)',
               '(x = x)',
               '(boy(x) | girl(x))',
               '(boy(x) & girl(x))',
               'love(adam, x)',
               'love(x, adam)',
               '-(x = adam)',
               'exists z22. love(x, z22)',
               'exists y. love(y, x)',
               'all y. (girl(y) -> love(x, y))',
               'all y. (girl(y) -> love(y, x))',
               'all y. (girl(y) -> (boy(x) & love(y, x)))',
               '(boy(x) & all y. (girl(y) -> love(x, y)))',
               '(boy(x) & all y. (girl(y) -> love(y, x)))',
               '(boy(x) & exists y. (girl(y) & love(y, x)))',
               '(girl(x) -> dog(x))',
               'all y. (dog(y) -> (x = y))',
               'exists y. love(y, x)',
               'exists y. (love(adam, y) & love(y, x))'
                ]

    if trace:
        print m2

    lp = LogicParser()
    for fmla in formulas:
        print fmla
        lp.parse(fmla)

    parsed = [lp.parse(fmla) for fmla in formulas]

    for p in parsed:
        g2.purge()
        print "The satisfiers of '%s' are: %s" % (p, m2.satisfiers(p, 'x', g2, trace))


def demo(num=0, trace=None):
    """
    Run exists demos.

     - num = 1: propositional logic demo
     - num = 2: first order model demo (only if trace is set)
     - num = 3: first order sentences demo
     - num = 4: satisfaction of open formulas demo
     - any other value: run all the demos

    :param trace: trace = 1, or trace = 2 for more verbose tracing
    """
    demos = {
        1: propdemo,
        2: folmodel,
        3: foldemo,
        4: satdemo}

    try:
        demos[num](trace=trace)
    except KeyError:
        for num in demos:
            demos[num](trace=trace)


if __name__ == "__main__":
    demo(2, trace=0)



# Natural Language Toolkit: Stemmer Interface
#
# Copyright (C) 2001-2012 NLTK Project
# Author: Trevor Cohn <tacohn@cs.mu.oz.au>
#         Edward Loper <edloper@gradient.cis.upenn.edu>
#         Steven Bird <sb@csse.unimelb.edu.au>
# URL: <http://www.nltk.org/>
# For license information, see LICENSE.TXT

class StemmerI(object):
    """
    A processing interface for removing morphological affixes from
    words.  This process is known as stemming.

    """
    def stem(self, token):
        """
        Strip affixes from the token and return the stem.

        :param token: The token that should be stemmed.
        :type token: str
        """
        raise NotImplementedError()


if __name__ == "__main__":
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



# Natural Language Toolkit: Language Models
#
# Copyright (C) 2001-2012 NLTK Project
# Authors: Steven Bird <sb@csse.unimelb.edu.au>
#          Daniel Blanchard <dblanchard@ets.org>
# URL: <http://www.nltk.org/>
# For license information, see LICENSE.TXT

from itertools import chain
from math import log

from nltk.probability import (ConditionalProbDist, ConditionalFreqDist,
                              SimpleGoodTuringProbDist)
from nltk.util import ingrams
from nltk.model.api import ModelI


def _estimator(fdist, bins):
    """
    Default estimator function using a SimpleGoodTuringProbDist.
    """
    # can't be an instance method of NgramModel as they
    # can't be pickled either.
    return SimpleGoodTuringProbDist(fdist)


class NgramModel(ModelI):
    """
    A processing interface for assigning a probability to the next word.
    """

    # add cutoff
    def __init__(self, n, train, pad_left=True, pad_right=False,
                 estimator=None, *estimator_args, **estimator_kwargs):
        """
        Create an ngram language model to capture patterns in n consecutive
        words of training text.  An estimator smooths the probabilities derived
        from the text and may allow generation of ngrams not seen during
        training.

            >>> from nltk.corpus import brown
            >>> from nltk.probability import LidstoneProbDist
            >>> est = lambda fdist, bins: LidstoneProbDist(fdist, 0.2)
            >>> lm = NgramModel(3, brown.words(categories='news'), estimator=est)
            >>> lm
            <NgramModel with 91603 3-grams>
            >>> lm._backoff
            <NgramModel with 62888 2-grams>
            >>> lm.entropy(['The', 'Fulton', 'County', 'Grand', 'Jury', 'said',
            ... 'Friday', 'an', 'investigation', 'of', "Atlanta's", 'recent',
            ... 'primary', 'election', 'produced', '``', 'no', 'evidence',
            ... "''", 'that', 'any', 'irregularities', 'took', 'place', '.'])
            ... # doctest: +ELLIPSIS
            0.5776...

        :param n: the order of the language model (ngram size)
        :type n: int
        :param train: the training text
        :type train: list(str) or list(list(str))
        :param pad_left: whether to pad the left of each sentence with an (n-1)-gram of empty strings
        :type pad_left: bool
        :param pad_right: whether to pad the right of each sentence with an (n-1)-gram of empty strings
        :type pad_right: bool
        :param estimator: a function for generating a probability distribution
        :type estimator: a function that takes a ConditionalFreqDist and
            returns a ConditionalProbDist
        :param estimator_args: Extra arguments for estimator.
            These arguments are usually used to specify extra
            properties for the probability distributions of individual
            conditions, such as the number of bins they contain.
            Note: For backward-compatibility, if no arguments are specified, the
            number of bins in the underlying ConditionalFreqDist are passed to
            the estimator as an argument.
        :type estimator_args: (any)
        :param estimator_kwargs: Extra keyword arguments for the estimator
        :type estimator_kwargs: (any)
        """

        # protection from cryptic behavior for calling programs
        # that use the pre-2.0.2 interface
        assert(isinstance(pad_left, bool))
        assert(isinstance(pad_right, bool))

        self._n = n
        self._lpad = ('',) * (n - 1) if pad_left else ()
        self._rpad = ('',) * (n - 1) if pad_right else ()

        if estimator is None:
            estimator = _estimator

        cfd = ConditionalFreqDist()
        self._ngrams = set()
        
            
        # If given a list of strings instead of a list of lists, create enclosing list
        if (train is not None) and isinstance(train[0], basestring):
            train = [train]

        for sent in train:
            for ngram in ingrams(chain(self._lpad, sent, self._rpad), n):
                self._ngrams.add(ngram)
                context = tuple(ngram[:-1])
                token = ngram[-1]
                cfd[context].inc(token)

        if not estimator_args and not estimator_kwargs:
            self._model = ConditionalProbDist(cfd, estimator, len(cfd))
        else:
            self._model = ConditionalProbDist(cfd, estimator, *estimator_args, **estimator_kwargs)

        # recursively construct the lower-order models
        if n > 1:
            self._backoff = NgramModel(n-1, train, pad_left, pad_right,
                                       estimator, *estimator_args, **estimator_kwargs)

    def prob(self, word, context):
        """
        Evaluate the probability of this word in this context using Katz Backoff.

        :param word: the word to get the probability of
        :type word: str
        :param context: the context the word is in
        :type context: list(str)
        """

        context = tuple(context)
        if (context + (word,) in self._ngrams) or (self._n == 1):
            return self[context].prob(word)
        else:
            return self._alpha(context) * self._backoff.prob(word, context[1:])

    def _alpha(self, tokens):
        return self._beta(tokens) / self._backoff._beta(tokens[1:])

    def _beta(self, tokens):
        if tokens in self:
            return self[tokens].discount()
        else:
            return 1

    def logprob(self, word, context):
        """
        Evaluate the (negative) log probability of this word in this context.

        :param word: the word to get the probability of
        :type word: str
        :param context: the context the word is in
        :type context: list(str)
        """

        return -log(self.prob(word, context), 2)

    def choose_random_word(self, context):
        '''
        Randomly select a word that is likely to appear in this context.

        :param context: the context the word is in
        :type context: list(str)
        '''

        return self.generate(1, context)[-1]

    # NB, this will always start with same word if the model
    # was trained on a single text
    def generate(self, num_words, context=()):
        '''
        Generate random text based on the language model.

        :param num_words: number of words to generate
        :type num_words: int
        :param context: initial words in generated string
        :type context: list(str)
        '''

        text = list(context)
        for i in range(num_words):
            text.append(self._generate_one(text))
        return text

    def _generate_one(self, context):
        context = (self._lpad + tuple(context))[-self._n+1:]
        # print "Context (%d): <%s>" % (self._n, ','.join(context))
        if context in self:
            return self[context].generate()
        elif self._n > 1:
            return self._backoff._generate_one(context[1:])
        else:
            return '.'

    def entropy(self, text):
        """
        Calculate the approximate cross-entropy of the n-gram model for a
        given evaluation text.
        This is the average log probability of each word in the text.

        :param text: words to use for evaluation
        :type text: list(str)
        """

        e = 0.0
        text = list(self._lpad) + text + list(self._rpad)
        for i in range(self._n-1, len(text)):
            context = tuple(text[i-self._n+1:i])
            token = text[i]
            e += self.logprob(token, context)
        return e / float(len(text) - (self._n-1))

    def perplexity(self, text):
        """
        Calculates the perplexity of the given text.
        This is simply 2 ** cross-entropy for the text.

        :param text: words to calculate perplexity of
        :type text: list(str)
        """

        return pow(2.0, self.entropy(text))

    def __contains__(self, item):
        return tuple(item) in self._model

    def __getitem__(self, item):
        return self._model[tuple(item)]

    def __repr__(self):
        return '<NgramModel with %d %d-grams>' % (len(self._ngrams), self._n)


if __name__ == "__main__":
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



# -*- coding: utf-8 -*-
# Natural Language Toolkit: Interface to the Stanford NER-tagger
#
# Copyright (C) 2001-2012 NLTK Project
# Author: Nitin Madnani <nmadnani@ets.org>
#         Rami Al-Rfou' <ralrfou@cs.stonybrook.edu>
# URL: <http://www.nltk.org/>
# For license information, see LICENSE.TXT

"""
A module for interfacing with the Stanford taggers.
"""

import os
import tempfile
from subprocess import PIPE

from nltk.internals import find_file, find_jar, config_java, java, _java_options
from nltk.tag.api import TaggerI

_stanford_url = 'http://nlp.stanford.edu/software'

class StanfordTagger(TaggerI):
    """
    An interface to Stanford taggers. Subclasses must define:

    - ``_cmd`` property: A property that returns the command that will be
      executed.
    - ``_SEPARATOR``: Class constant that represents that character that
      is used to separate the tokens from their tags.
    - ``_JAR`` file: Class constant that represents the jar file name.
    """

    _SEPARATOR = ''
    _JAR = ''

    def __init__(self, path_to_model, path_to_jar=None, encoding=None, verbose=False, java_options='-mx1000m'):

        self._stanford_jar = find_jar(
                self._JAR, path_to_jar,
                searchpath=(), url=_stanford_url,
                verbose=verbose)

        self._stanford_model = find_file(path_to_model,
                env_vars=('STANFORD_MODELS'), verbose=verbose)
        self._encoding = encoding
        self.java_options = java_options

    @property
    def _cmd(self):
      raise NotImplementedError

    def tag(self, tokens):
        return self.batch_tag([tokens])[0]

    def batch_tag(self, sentences):
        encoding = self._encoding
        default_options = ' '.join(_java_options)
        config_java(options=self.java_options, verbose=False)

        # Create a temporary input file
        _input_fh, self._input_file_path = tempfile.mkstemp(text=True)

        if encoding:
            self._cmd.extend(['-encoding', encoding])

        # Write the actual sentences to the temporary input file
        _input_fh = os.fdopen(_input_fh, 'w')
        _input = '\n'.join((' '.join(x) for x in sentences))
        if isinstance(_input, unicode) and encoding:
            _input = _input.encode(encoding)
        _input_fh.write(_input)
        _input_fh.close()

        # Run the tagger and get the output
        stanpos_output, _stderr = java(self._cmd,classpath=self._stanford_jar, \
                                                       stdout=PIPE, stderr=PIPE)
        if encoding:
            stanpos_output = stanpos_output.decode(encoding)

        # Delete the temporary file
        os.unlink(self._input_file_path)

        # Return java configurations to their default values
        config_java(options=default_options, verbose=False)

        return self.parse_output(stanpos_output)

    def parse_output(self, text):
        # Output the tagged sentences
        tagged_sentences = []
        for tagged_sentence in text.strip().split("\n"):
            sentence = []
            for tagged_word in tagged_sentence.strip().split():
                word_tags = tagged_word.strip().split(self._SEPARATOR)
                sentence.append((''.join(word_tags[:-1]), word_tags[-1]))
            tagged_sentences.append(sentence)
        return tagged_sentences

class POSTagger(StanfordTagger):
    """
    A class for pos tagging with Stanford Tagger. The input is the paths to:
     - a model trained on training data
     - (optionally) the path to the stanford tagger jar file. If not specified here,
       then this jar file must be specified in the CLASSPATH envinroment variable.
     - (optionally) the encoding of the training data (default: ASCII)

    Example:

    .. doctest::
        :options: +SKIP

        >>> from nltk.tag.stanford import POSTagger
        >>> st = POSTagger('/usr/share/stanford-postagger/models/english-bidirectional-distsim.tagger',
        ...                '/usr/share/stanford-postagger/stanford-postagger.jar')
        >>> st.tag('What is the airspeed of an unladen swallow ?'.split())
        [('What', 'WP'), ('is', 'VBZ'), ('the', 'DT'), ('airspeed', 'NN'), ('of', 'IN'), ('an', 'DT'), ('unladen', 'JJ'), ('swallow', 'VB'), ('?', '.')]
    """

    _SEPARATOR = '_'
    _JAR = 'stanford-postagger.jar'

    def __init__(self, *args, **kwargs):
        super(POSTagger, self).__init__(*args, **kwargs)

    @property
    def _cmd(self):
        return ['edu.stanford.nlp.tagger.maxent.MaxentTagger', \
                '-model', self._stanford_model, '-textFile', \
                self._input_file_path, '-tokenize', 'false']

class NERTagger(StanfordTagger):
    """
    A class for ner tagging with Stanford Tagger. The input is the paths to:

    - a model trained on training data
    - (optionally) the path to the stanford tagger jar file. If not specified here,
      then this jar file must be specified in the CLASSPATH envinroment variable.
    - (optionally) the encoding of the training data (default: ASCII)

    Example:

    .. doctest::
        :options: +SKIP

        >>> from nltk.tag.stanford import NERTagger
        >>> st = NERTagger('/usr/share/stanford-ner/classifiers/all.3class.distsim.crf.ser.gz',
        ...                '/usr/share/stanford-ner/stanford-ner.jar')
        >>> st.tag('Rami Eid is studying at Stony Brook University in NY'.split())
        [('Rami', 'PERSON'), ('Eid', 'PERSON'), ('is', 'O'), ('studying', 'O'),
         ('at', 'O'), ('Stony', 'ORGANIZATION'), ('Brook', 'ORGANIZATION'),
         ('University', 'ORGANIZATION'), ('in', 'O'), ('NY', 'LOCATION')]
    """

    _SEPARATOR = '/'
    _JAR = 'stanford-ner.jar'
    _FORMAT = 'slashTags'

    def __init__(self, *args, **kwargs):
        super(NERTagger, self).__init__(*args, **kwargs)

    @property
    def _cmd(self):
        return ['edu.stanford.nlp.ie.crf.CRFClassifier', \
                '-loadClassifier', self._stanford_model, '-textFile', \
                self._input_file_path, '-outputFormat', self._FORMAT]

    def parse_output(self, text):
      if self._FORMAT == 'slashTags':
        return super(NERTagger, self).parse_output(text)
      raise NotImplementedError


if __name__ == "__main__":
    import doctest
    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



# Natural Language Toolkit: Ngram Association Measures
#
# Copyright (C) 2001-2012 NLTK Project
# Author: Joel Nothman <jnothman@student.usyd.edu.au>
# URL: <http://nltk.org>
# For license information, see LICENSE.TXT

"""
Provides scoring functions for a number of association measures through a
generic, abstract implementation in ``NgramAssocMeasures``, and n-specific
``BigramAssocMeasures`` and ``TrigramAssocMeasures``.
"""

import math as _math
_log2 = lambda x: _math.log(x, 2.0)
_ln = _math.log

_product = lambda s: reduce(lambda x, y: x * y, s)

_SMALL = 1e-20


### Indices to marginals arguments:

NGRAM = 0
"""Marginals index for the ngram count"""

UNIGRAMS = -2
"""Marginals index for a tuple of each unigram count"""

TOTAL = -1
"""Marginals index for the number of words in the data"""


class NgramAssocMeasures(object):
    """
    An abstract class defining a collection of generic association measures.
    Each public method returns a score, taking the following arguments::

        score_fn(count_of_ngram,
                 (count_of_n-1gram_1, ..., count_of_n-1gram_j),
                 (count_of_n-2gram_1, ..., count_of_n-2gram_k),
                 ...,
                 (count_of_1gram_1, ..., count_of_1gram_n),
                 count_of_total_words)

    See ``BigramAssocMeasures`` and ``TrigramAssocMeasures``

    Inheriting classes should define a property _n, and a method _contingency
    which calculates contingency values from marginals in order for all
    association measures defined here to be usable.
    """

    @staticmethod
    def _contingency(*marginals):
        """Calculates values of a contingency table from marginal values."""
        raise NotImplementedError, ("The contingency table is not available"
                                    "in the general ngram case")

    @staticmethod
    def _marginals(*contingency):
        """Calculates values of contingency table marginals from its values."""
        raise NotImplementedError, ("The contingency table is not available"
                                    "in the general ngram case")

    @classmethod
    def _expected_values(cls, cont):
        """Calculates expected values for a contingency table."""
        n_all = sum(cont)
        bits = [1 << i for i in range(cls._n)]

        # For each contingency table cell
        for i in range(len(cont)):
            # Yield the expected value
            yield (_product(cont[i] + cont[i ^ j] for j in bits) /
                   float(n_all ** 2))

    @staticmethod
    def raw_freq(*marginals):
        """Scores ngrams by their frequency"""
        return float(marginals[NGRAM]) / marginals[TOTAL]

    @classmethod
    def student_t(cls, *marginals):
        """Scores ngrams using Student's t test with independence hypothesis
        for unigrams, as in Manning and Schutze 5.3.2.
        """
        return ((marginals[NGRAM] * marginals[TOTAL] -
                 _product(marginals[UNIGRAMS])) /
                (marginals[TOTAL] ** (cls._n - 1) *
                (marginals[NGRAM] + _SMALL) ** .5))

    @classmethod
    def chi_sq(cls, *marginals):
        """Scores ngrams using Pearson's chi-square as in Manning and Schutze
        5.3.3.
        """
        cont = cls._contingency(*marginals)
        exps = cls._expected_values(cont)
        return sum((obs - exp) ** 2 / (exp + _SMALL)
                   for obs, exp in zip(cont, exps))

    @staticmethod
    def mi_like(*marginals, **kwargs):
        """Scores ngrams using a variant of mutual information. The keyword
        argument power sets an exponent (default 3) for the numerator. No
        logarithm of the result is calculated.
        """
        return (marginals[NGRAM] ** kwargs.get('power', 3) /
                float(_product(marginals[UNIGRAMS])))

    @classmethod
    def pmi(cls, *marginals):
        """Scores ngrams by pointwise mutual information, as in Manning and
        Schutze 5.4.
        """
        return (_log2(marginals[NGRAM] * marginals[TOTAL] ** (cls._n - 1)) -
                _log2(_product(marginals[UNIGRAMS])))

    @classmethod
    def likelihood_ratio(cls, *marginals):
        """Scores ngrams using likelihood ratios as in Manning and Schutze 5.3.4.
        """
        cont = cls._contingency(*marginals)
        return (cls._n * 2 *
                sum(obs * _ln(float(obs) / (exp + _SMALL) + _SMALL)
                    for obs, exp in zip(cont, cls._expected_values(cont))))

    @classmethod
    def poisson_stirling(cls, *marginals):
        """Scores ngrams using the Poisson-Stirling measure."""
        exp = (_product(marginals[UNIGRAMS]) /
              float(marginals[TOTAL] ** (cls._n - 1)))
        return marginals[NGRAM] * (_log2(marginals[NGRAM] / exp) - 1)

    @classmethod
    def jaccard(cls, *marginals):
        """Scores ngrams using the Jaccard index."""
        cont = cls._contingency(*marginals)
        return float(cont[0]) / sum(cont[:-1])


class BigramAssocMeasures(NgramAssocMeasures):
    """
    A collection of trigram association measures. Each association measure
    is provided as a function with three arguments::

        bigram_score_fn(n_ii, (n_ix, n_xi), n_xx)

    The arguments constitute the marginals of a contingency table, counting
    the occurrences of particular events in a corpus. The letter i in the
    suffix refers to the appearance of the word in question, while x indicates
    the appearance of any word. Thus, for example::

        n_ii counts (w1, w2), i.e. the bigram being scored
        n_ix counts (w1, *)
        n_xi counts (*, w2)
        n_xx counts (*, *), i.e. any bigram

    This may be shown with respect to a contingency table::

                w1    ~w1
             ------ ------
         w2 | n_ii | n_oi | = n_xi
             ------ ------
        ~w2 | n_io | n_oo |
             ------ ------
             = n_ix        TOTAL = n_xx
    """

    _n = 2

    @staticmethod
    def _contingency(n_ii, (n_ix, n_xi), n_xx):
        """Calculates values of a bigram contingency table from marginal values."""
        n_oi = n_xi - n_ii
        n_io = n_ix - n_ii
        return (n_ii, n_oi, n_io, n_xx - n_ii - n_oi - n_io)

    @staticmethod
    def _marginals(n_ii, n_oi, n_io, n_oo):
        """Calculates values of contingency table marginals from its values."""
        return (n_ii, (n_oi + n_ii, n_io + n_ii), n_oo + n_oi + n_io + n_ii)

    @staticmethod
    def _expected_values(cont):
        """Calculates expected values for a contingency table."""
        n_xx = sum(cont)
        # For each contingency table cell
        for i in range(4):
            yield (cont[i] + cont[i ^ 1]) * (cont[i] + cont[i ^ 2]) / float(n_xx)

    @classmethod
    def phi_sq(cls, *marginals):
        """Scores bigrams using phi-square, the square of the Pearson correlation
        coefficient.
        """
        n_ii, n_io, n_oi, n_oo = cls._contingency(*marginals)

        return (float((n_ii*n_oo - n_io*n_oi)**2) /
                ((n_ii + n_io) * (n_ii + n_oi) * (n_io + n_oo) * (n_oi + n_oo)))

    @classmethod
    def chi_sq(cls, n_ii, (n_ix, n_xi), n_xx):
        """Scores bigrams using chi-square, i.e. phi-sq multiplied by the number
        of bigrams, as in Manning and Schutze 5.3.3.
        """
        return n_xx * cls.phi_sq(n_ii, (n_ix, n_xi), n_xx)

    @staticmethod
    def dice(n_ii, (n_ix, n_xi), n_xx):
        """Scores bigrams using Dice's coefficient."""
        return 2 * float(n_ii) / (n_ix + n_xi)


class TrigramAssocMeasures(NgramAssocMeasures):
    """
    A collection of trigram association measures. Each association measure
    is provided as a function with four arguments::

        trigram_score_fn(n_iii,
                         (n_iix, n_ixi, n_xii),
                         (n_ixx, n_xix, n_xxi),
                         n_xxx)

    The arguments constitute the marginals of a contingency table, counting
    the occurrences of particular events in a corpus. The letter i in the
    suffix refers to the appearance of the word in question, while x indicates
    the appearance of any word. Thus, for example:
    n_iii counts (w1, w2, w3), i.e. the trigram being scored
    n_ixx counts (w1, *, *)
    n_xxx counts (*, *, *), i.e. any trigram
    """

    _n = 3

    @staticmethod
    def _contingency(n_iii,
                 (n_iix, n_ixi, n_xii),
                 (n_ixx, n_xix, n_xxi),
                 n_xxx):
        """Calculates values of a trigram contingency table (or cube) from marginal
        values.
        """
        n_oii = n_xii - n_iii
        n_ioi = n_ixi - n_iii
        n_iio = n_iix - n_iii
        n_ooi = n_xxi - n_iii - n_oii - n_ioi
        n_oio = n_xix - n_iii - n_oii - n_iio
        n_ioo = n_ixx - n_iii - n_ioi - n_iio
        n_ooo = n_xxx - n_iii - n_oii - n_ioi - n_iio - n_ooi - n_oio - n_ioo

        return (n_iii, n_oii, n_ioi, n_ooi,
                n_iio, n_oio, n_ioo, n_ooo)

    @staticmethod
    def _marginals(*contingency):
        """Calculates values of contingency table marginals from its values."""
        n_iii, n_oii, n_ioi, n_ooi, n_iio, n_oio, n_ioo, n_ooo = contingency
        return (n_iii,
                (n_iii + n_iio, n_iii + n_ioi, n_iii + n_oii),
                (n_iii + n_ioi + n_iio + n_ioo,
                 n_iii + n_oii + n_iio + n_oio,
                 n_iii + n_oii + n_ioi + n_ooi),
                sum(contingency))


class ContingencyMeasures(object):
    """Wraps NgramAssocMeasures classes such that the arguments of association
    measures are contingency table values rather than marginals.
    """

    def __init__(self, measures):
        """Constructs a ContingencyMeasures given a NgramAssocMeasures class"""
        self.__class__.__name__ = 'Contingency' + measures.__class__.__name__
        for k in dir(measures):
            if k.startswith('__'):
                continue
            v = getattr(measures, k)
            if not k.startswith('_'):
                v = self._make_contingency_fn(measures, v)
            setattr(self, k, v)

    @staticmethod
    def _make_contingency_fn(measures, old_fn):
        """From an association measure function, produces a new function which
        accepts contingency table values as its arguments.
        """
        def res(*contingency):
            return old_fn(*measures._marginals(*contingency))
        res.__doc__ = old_fn.__doc__
        res.__name__ = old_fn.__name__
        return res




# Natural Language Toolkit: Toolbox Reader
#
# Copyright (C) 2001-2012 NLTK Project
# Author: Greg Aumann <greg_aumann@sil.org>
#         Stuart Robinson <Stuart.Robinson@mpi.nl>
#         Steven Bird <sb@csse.unimelb.edu.au>
# URL: <http://www.nltk.org/>
# For license information, see LICENSE.TXT

"""
Module for reading, writing and manipulating
Toolbox databases and settings fileids.
"""

import os
import re
import codecs

from nltk.toolbox import ToolboxData

from util import *
from api import *

class ToolboxCorpusReader(CorpusReader):
    def xml(self, fileids, key=None):
        return concat([ToolboxData(path, enc).parse(key)
                       for (path, enc) in self.abspaths(fileids, True)])

    def fields(self, fileids, strip=True, unwrap=True, encoding=None,
               errors='strict', unicode_fields=None):
        return concat([list(ToolboxData(fileid,enc).fields(
                             strip, unwrap, encoding, errors, unicode_fields))
                       for (fileid, enc)
                       in self.abspaths(fileids, include_encoding=True)])

    # should probably be done lazily:
    def entries(self, fileids, **kwargs):
        if 'key' in kwargs:
            key = kwargs['key']
            del kwargs['key']
        else:
            key = 'lx'  # the default key in MDF
        entries = []
        for marker, contents in self.fields(fileids, **kwargs):
            if marker == key:
                entries.append((contents, []))
            else:
                try:
                    entries[-1][-1].append((marker, contents))
                except IndexError:
                    pass
        return entries

    def words(self, fileids, key='lx'):
        return [contents for marker, contents in self.fields(fileids) if marker == key]

    def raw(self, fileids):
        if fileids is None: fileids = self._fileids
        elif isinstance(fileids, basestring): fileids = [fileids]
        return concat([self.open(f).read() for f in fileids])


def demo():
    pass

if __name__ == '__main__':
    demo()
